# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import json
import os
import sqlite3
import sys
from contextlib import suppress
from typing import Dict, Iterable, Optional, List, Union

from cylc.flow.exceptions import InputError
from cylc.flow.cycling.util import add_offset
from cylc.flow.cycling.integer import (
    IntegerPoint,
    IntegerInterval
)
from cylc.flow.flow_mgr import stringify_flow_nums
from cylc.flow.pathutil import expand_path
from cylc.flow.rundb import CylcWorkflowDAO
from cylc.flow.task_outputs import (
    TASK_OUTPUT_SUCCEEDED,
    TASK_OUTPUT_FAILED,
    TASK_OUTPUT_FINISHED,
)
from cylc.flow.util import deserialise_set
from metomi.isodatetime.parsers import TimePointParser
from metomi.isodatetime.exceptions import ISO8601SyntaxError


output_fallback_msg = (
    "Unable to filter by task output label for tasks run in Cylc versions "
    "between 8.0.0-8.3.0. Falling back to filtering by task message instead."
)


class CylcWorkflowDBChecker:
    """Object for querying task status or outputs from a workflow database.

    Back-compat and task outputs:
        # Cylc 7 stored {trigger: message} for custom outputs only.
        1|foo|{"x": "the quick brown"}

        # Cylc 8 (pre-8.3.0) stored [message] only, for all outputs.
        1|foo|[1]|["submitted", "started", "succeeded", "the quick brown"]

        # Cylc 8 (8.3.0+) stores {trigger: message} for all ouputs.
        1|foo|[1]|{"submitted": "submitted", "started": "started",
                   "succeeded": "succeeded", "x": "the quick brown"}
    """

    def __init__(self, rund, workflow, db_path=None):
        # (Explicit dp_path arg is to make testing easier).
        if db_path is None:
            # Infer DB path from workflow name and run dir.
            db_path = expand_path(
                rund, workflow, "log", CylcWorkflowDAO.DB_FILE_BASE_NAME
            )
        if not os.path.exists(db_path):
            raise OSError(errno.ENOENT, os.strerror(errno.ENOENT), db_path)

        self.conn: sqlite3.Connection = sqlite3.connect(db_path, timeout=10.0)

        # Get workflow point format.
        try:
            self.db_point_fmt = self._get_db_point_format()
            self.c7_back_compat_mode = False
        except sqlite3.OperationalError as exc:
            # BACK COMPAT: Cylc 7 DB (see method below).
            try:
                self.db_point_fmt = self._get_db_point_format_compat()
                self.c7_back_compat_mode = True
            except sqlite3.OperationalError:
                with suppress(Exception):
                    self.conn.close()
                raise exc  # original error

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Close DB connection when leaving context manager."""
        self.conn.close()

    def adjust_point_to_db(self, cycle, offset):
        """Adjust a cycle point (with offset) to the DB point format.

        Cycle point queries have to match in the DB as string literals,
        so we convert given cycle points (e.g., from the command line)
        to the DB point format before making the query.

        """
        if cycle is None or "*" in cycle:
            if offset is not None:
                raise InputError(
                    f'Cycle point "{cycle}" is not compatible with an offset.'
                )
            # Nothing to do
            return cycle

        if offset is not None:
            if self.db_point_fmt is None:
                # integer cycling
                cycle = str(
                    IntegerPoint(cycle) +
                    IntegerInterval(offset)
                )
            else:
                cycle = str(
                    add_offset(cycle, offset)
                )

        if self.db_point_fmt is None:
            return cycle

        # Convert cycle point to DB format.
        try:
            cycle = str(
                TimePointParser().parse(
                    cycle, dump_format=self.db_point_fmt
                )
            )
        except ISO8601SyntaxError:
            raise InputError(
                f'Cycle point "{cycle}" is not compatible'
                f' with DB point format "{self.db_point_fmt}"'
            )
        return cycle

    @staticmethod
    def display_maps(res, old_format=False, pretty_print=False):
        if not res:
            sys.stderr.write("INFO: No results to display.\n")
        else:
            for row in res:
                if old_format:
                    sys.stdout.write(', '.join(row) + '\n')
                else:
                    out = f"{row[1]}/{row[0]}:"  # cycle/task:
                    status_or_outputs = row[2]
                    if pretty_print:
                        with suppress(json.decoder.JSONDecodeError):
                            status_or_outputs = (
                                json.dumps(
                                    json.loads(
                                        status_or_outputs.replace("'", '"')
                                    ),
                                    indent=4
                                )
                            )
                    out += status_or_outputs
                    if len(row) == 4:
                        out += row[3]  # flow
                    sys.stdout.write(out + "\n")

    def _get_db_point_format(self):
        """Query a workflow database for a 'cycle point format' entry"""
        for row in self.conn.execute(
            rf'''
                SELECT
                    value
                FROM
                    {CylcWorkflowDAO.TABLE_WORKFLOW_PARAMS}
                WHERE
                    key==?
            ''',  # nosec (table name is code constant)
            ['cycle_point_format']
        ):
            return row[0]

    def _get_db_point_format_compat(self):
        """Query a Cylc 7 suite database for 'cycle point format'."""
        # BACK COMPAT: Cylc 7 DB
        # Workflows parameters table name change.
        # from:
        #    8.0.x
        # to:
        #    8.1.x
        # remove at:
        #    8.x
        for row in self.conn.execute(
            rf'''
                SELECT
                    value
                FROM
                    {CylcWorkflowDAO.TABLE_SUITE_PARAMS}
                WHERE
                    key==?
            ''',  # nosec (table name is code constant)
            ['cycle_point_format']
        ):
            return row[0]

    def workflow_state_query(
        self,
        task: Optional[str] = None,
        cycle: Optional[str] = None,
        selector: Optional[str] = None,
        is_trigger: Optional[bool] = False,
        is_message: Optional[bool] = False,
        flow_num: Optional[int] = None,
        print_outputs: bool = False
    ) -> List[List[str]]:
        """Query task status or outputs (by trigger or message) in a database.

        Args:
            task:
                task name
            cycle:
                cycle point
            selector:
                task status, trigger name, or message
            is_trigger:
                intpret the selector as a trigger
            is_message:
                interpret the selector as a task message

        Return:
            A list of results for all tasks that match the query.
        [
            [name, cycle, result, [flow]],
            ...
        ]

        "result" is single string:
          - for status queries: the task status
          - for output queries: a serialized dict of completed outputs
            {trigger: message}

        """
        stmt_args = []
        stmt_wheres = []

        if is_trigger or is_message:
            target_table = CylcWorkflowDAO.TABLE_TASK_OUTPUTS
            mask = "name, cycle, outputs"
        else:
            target_table = CylcWorkflowDAO.TABLE_TASK_STATES
            mask = "name, cycle, status"

        if not self.c7_back_compat_mode:
            # Cylc 8 DBs only
            mask += ", flow_nums"

        stmt = rf'''
            SELECT
                {mask}
            FROM
                {target_table}
        '''  # nosec
        # * mask is hardcoded
        # * target_table is a code constant

        # Select from DB by name, cycle, status.
        # (Outputs and flow_nums are serialised).
        if task:
            if '*' in task:
                # Replace Cylc ID wildcard with Sqlite query wildcard.
                task = task.replace('*', '%')
                stmt_wheres.append("name like ?")
            else:
                stmt_wheres.append("name==?")
            stmt_args.append(task)

        if cycle:
            if '*' in cycle:
                # Replace Cylc ID wildcard with Sqlite query wildcard.
                cycle = cycle.replace('*', '%')
                stmt_wheres.append("cycle like ?")
            else:
                stmt_wheres.append("cycle==?")
            stmt_args.append(cycle)

        if (
            selector is not None
            and target_table == CylcWorkflowDAO.TABLE_TASK_STATES
        ):
            # Can select by status in the DB but not outputs.
            stmt_wheres.append("status==?")
            stmt_args.append(selector)

        if stmt_wheres:
            stmt += "WHERE\n    " + (" AND ").join(stmt_wheres)

        if target_table == CylcWorkflowDAO.TABLE_TASK_STATES:
            # (outputs table doesn't record submit number)
            stmt += r"ORDER BY submit_num"

        # Query the DB and drop incompatible rows.
        db_res = []
        for row in self.conn.execute(stmt, stmt_args):
            # name, cycle, status_or_outputs, [flow_nums]
            res = list(row[:3])
            if row[2] is None:
                # status can be None in Cylc 7 DBs
                continue
            if not self.c7_back_compat_mode:
                flow_nums = deserialise_set(row[3])
                if flow_num is not None and flow_num not in flow_nums:
                    # skip result, wrong flow
                    continue
                fstr = stringify_flow_nums(flow_nums)
                if fstr:
                    res.append(fstr)
            db_res.append(res)

        if target_table == CylcWorkflowDAO.TABLE_TASK_STATES:
            return db_res

        warn_output_fallback = is_trigger
        results = []
        for row in db_res:
            outputs: Union[Dict[str, str], List[str]] = json.loads(row[2])
            if isinstance(outputs, dict):
                messages: Iterable[str] = outputs.values()
            else:
                # Cylc 8 pre 8.3.0 back-compat: list of output messages
                messages = outputs
                if warn_output_fallback:
                    print(f"WARNING - {output_fallback_msg}", file=sys.stderr)
                    warn_output_fallback = False
            if (
                selector is None or
                (is_message and selector in messages) or
                (is_trigger and self._selector_in_outputs(selector, outputs))
            ):
                results.append(row[:2] + [str(outputs)] + row[3:])

        return results

    @staticmethod
    def _selector_in_outputs(selector: str, outputs: Iterable[str]) -> bool:
        """Check if a selector, including "finished", is in the outputs.

        Examples:
            >>> this = CylcWorkflowDBChecker._selector_in_outputs
            >>> this('moop', ['started', 'moop'])
            True
            >>> this('moop', ['started'])
            False
            >>> this('finished', ['succeeded'])
            True
            >>> this('finish', ['failed'])
            True
        """
        return selector in outputs or (
            selector in (TASK_OUTPUT_FINISHED, "finish")
            and (
                TASK_OUTPUT_SUCCEEDED in outputs
                or TASK_OUTPUT_FAILED in outputs
            )
        )



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""cylc scan [OPTIONS]

List your source, and/or installed, and/or running workflows.

By default, running workflows are listed as indicated by the presence of
scheduler contact files, ~/cylc-run/<Workflow-ID>/.service/contact.

With "--ping" or "-t rich", attempt to contact the schedulers. If any are not
found to be running, their contact files will be removed (these files may left
behind if the scheduler did not shut down cleanly).

Examples:
  # list all active workflows (i.e. running or paused)
  $ cylc scan

  # show more information about active workflows
  $ cylc scan -t rich

  # don't rely on colour for job state totals
  $ cylc scan -t rich --colour-blind

  # list workflows that are installed but stopped or not run yet
  $ cylc scan --state stopped

  # list all workflows (active or inactive)
  $ cylc scan --state=running,paused,stopped
  $ cylc scan --state=all  # or using the shorthand

  # filter workflows by name
  $ cylc scan --name '^f.*'  # show only flows starting with "f"

  # list source workflows in tree format
  # (looks in the dirs configured by "global.cylc[install]source dirs")
  $ cylc scan --source -t tree

  # print contact file data in JSON format
  $ cylc scan -t json
"""

import asyncio
import json
from pathlib import Path
from typing import Callable, Optional, TYPE_CHECKING

from ansimarkup import ansiprint as cprint

from cylc.flow import LOG
from cylc.flow.cfgspec.glbl_cfg import glbl_cfg
from cylc.flow.exceptions import InputError
from cylc.flow.network.scan import (
    contact_info,
    filter_name,
    graphql_query,
    is_active,
    scan,
    scan_multi,
)
from cylc.flow.option_parsers import (
    CylcOptionParser as COP,
    Options
)
from cylc.flow.print_tree import get_tree
from cylc.flow.terminal import cli_function, DIM
from cylc.flow.util import natural_sort_key
from cylc.flow.workflow_files import ContactFileFields as Cont

if TYPE_CHECKING:
    from optparse import Values

# all supported workflow states
FLOW_STATES = {
    'running',
    'paused',
    'stopping',
    'stopped'
}


# workflow status colours
FLOW_STATE_CMAP = {
    # workflow state: term colour
    'running': 'green',
    'paused': 'fg 172',
    'stopping': 'fg 201',
    'stopped': 'red'
}


# workflow status symbols
FLOW_STATE_SYMBOLS = {
    # NOTE: the standard media control characters ▶️,, ⏸️,, ⏹️
    #       can appear wildly different font-depending and may not
    #       even be monospace
    'running': '▶',
    'paused': '‖',
    'stopping': '◧',
    'stopped': '■'
}


# document the flow states
__doc__ += '\n    '.join(
    [
        '\nWorkflow States:'
    ] + [
        (
            f'<{FLOW_STATE_CMAP[state]}>'
            f'{FLOW_STATE_SYMBOLS[state]}'
            f' {state}'
            f'</{FLOW_STATE_CMAP[state]}>'
        )
        for state in FLOW_STATES
    ]
)


# job icon colours
JOB_COLOURS = {
    # job status: term colour
    'submitted': 'fg 44',
    'submit-failed': 'fg 13',
    'running': 'fg 32',
    'succeeded': 'fg 35',
    'failed': 'fg 124'
}


# document the task states
__doc__ += '\n    '.join(
    [
        '\n\nTask States:'
    ] + [
        f'<{tag}>■ {state}</{tag}>'
        for state, tag in JOB_COLOURS.items()
    ]
)


# graphql fields to request from the workflow for the "rich" format
RICH_FIELDS = {
    'status': None,
    'stateTotals': None,
    'cylcVersion': None,
    'meta': {
        'title',
        'description',
    },

}


BAD_CONTACT_FILE_MSG = "Bad contact file: {flow_name}"


def get_option_parser() -> COP:
    """CLI opts for "cylc scan"."""
    parser = COP(
        __doc__, comms=True
    )

    parser.add_option(
        '--name', '-n',
        help=(
            'Filter workflows by registered name using a regex.'
            ' Can be used multiple times, workflows will be displayed if'
            ' their name matches ANY of the provided regexes.'
        ),
        action='append'
    )

    parser.add_option(
        '--states',
        help=(
            'Choose which workflows to display by providing a list of states'
            ' or "all" to show everything. See the full `cylc scan` help'
            ' for a list of supported states.'
        ),
        default='running,paused,stopping',
        action='store'
    )

    parser.add_option(
        '--source', '-s',
        help=(
            'List source workflows from configured source'
            ' directories (overrides the --states option).'
        ),
        default=False,
        action='store_true'
    )

    parser.add_option(
        '--format', '-t',
        help=(
            r'Output data and format (default "%default").'
            ' ("name": list the workflow IDs only)'
            ' ("plain": name,host:port,PID on one line)'
            ' ("tree": name,host:port,PID in tree format)'
            ' ("json": full contact data in JSON format)'
            ' ("rich": include task state summary data)'
        ),
        choices=('rich', 'plain', 'json', 'tree', 'name'),
        default='plain'
    )

    parser.add_option(
        '--sort',
        help='Sort flows by name before writing them (takes longer).',
        action='store_true'
    )

    parser.add_option(
        '--colour-blind', '--color-blind',
        help=(
            "Don't depend on colour to convey information. "
            'Use this rather than --color=never so you still get bold text.'
        ),
        action='store_true'
    )

    parser.add_option(
        '--ping',
        help=(
            "Connect to schedulers and remove contact files if they are not "
            "found to be running. Contact files can be left behind when "
            "schedulers get killed."
        ),
        action='store_true'
    )

    return parser


def state_totals(totals, colour_blind=False):
    """Return a string with a visual representation of workflow state totals.

    Args:
        totals (dict):
            State totals dictionary.
        colour_blind (bool):
            If True then we will not depend on colour to convey information.

    Returns:
        str

    """
    ret = []
    for state, tag in JOB_COLOURS.items():
        number = totals[state]
        if number == 0:
            continue
        if colour_blind:
            ret.append(f'<{tag}>{state}:{number}</{tag}>')
        else:
            ret.append(f'<{tag}>{number} ■</{tag}>')
    return ' '.join(ret)


def state_totals_key():
    """Return a key to accompany state_totals."""
    return '<b>Job State Key: </b>' + ' '.join(
        f'<{tag}>\u25A0 {name}</{tag}>'
        for name, tag in JOB_COLOURS.items()
    )


def _format_json(items, _):
    """A JSON formatter."""
    return json.dumps(
        [
            {
                key: str(value) if isinstance(value, Path) else value
                for key, value in flow.items()
            }
            for flow in items
        ],
        indent=4,
        sort_keys=True
    )


def _format_plain(flow, _):
    """A single line format of the form: <name> [<host>:<port>]"""
    if flow.get('contact'):
        try:
            return (
                f"<b>{flow['name']}</b> "
                f"{flow[Cont.HOST]}:{flow[Cont.PORT]} "
                f"{flow[Cont.PID]}"
            )
        except KeyError:
            LOG.warning(BAD_CONTACT_FILE_MSG.format(flow_name=flow['name']))
            return None
    else:
        return f'<{DIM}><b>{flow["name"]}</b></{DIM}>'


def _format_name_only(flow, _):
    """A single line format of the form: <name> [<host>:<port>]"""
    return flow['name']


def _format_rich(flow, opts):
    """A multiline format which pulls in metadata."""
    status = flow.get('status', 'stopped')
    if opts.colour_blind:
        name = f'{flow["name"]} ({status})'
    else:
        symbol = FLOW_STATE_SYMBOLS[status]
        tag = FLOW_STATE_CMAP[status]
        name = f'<{tag}>{symbol}</{tag}> {flow["name"]}'
    if not flow.get('contact') or 'status' not in flow:
        ret = [f'<{DIM}><b>{name}</b></{DIM}>']
    else:
        ret = [f'<b>{name}</b>']

        display = {
            'state totals': state_totals(
                flow['stateTotals'], opts.colour_blind
            ),
            **{
                key: flow['meta'][key] or f'<{DIM}>*null*</{DIM}>'
                for key in (
                    'title',
                    'description'
                )
            },
            **{
                name: flow[key]
                for name, key in (
                    ('version', 'cylcVersion'),
                    ('host', Cont.HOST),
                    ('port', Cont.PORT),
                    ('PID', Cont.PID)
                )
            }
        }
        maxlen = max(len(key) for key in display)
        for key, value in display.items():
            # format multiline strings by whitespace padding the lines
            value = ('\n' + (' ' * (maxlen + 7))).join(value.splitlines())
            # write out the key: value pairs
            ret.append(f'    {key: <{maxlen}}   {value}')
    return '\n'.join(ret)


def sort_function(flow):
    if (
        flow.get('status') == 'running'
        or flow.get('status') == 'paused'
    ):
        state = 0
    elif flow.get('contact'):
        state = 2
    else:
        state = 3
    return (state, *natural_sort_key(flow['name']))


async def _sorted(pipe, formatter, opts, write):
    """List all flows, sort, then print them individually."""
    ret = []
    async for item in pipe:
        ret.append(item)
    for flow in sorted(ret, key=sort_function):
        out = formatter(flow, opts)
        if out is not None:
            write(out)


async def _serial(pipe, formatter, opts, write):
    """List all flows, then print them as one."""
    ret = []
    async for item in pipe:
        ret.append(item)
    out = formatter(ret, opts)
    if out is not None:
        write(out)


async def _async(pipe, formatter, opts, write):
    """List and print flows individually."""
    async for flow in pipe:
        out = formatter(flow, opts)
        if out is not None:
            write(out)


async def _tree(pipe, formatter, opts, write):
    """List all flows, sort, then print them as a tree."""
    # get list of flows
    ret = []
    async for flow in pipe:
        ret.append(flow)

    tree = {}
    _construct_tree(ret, tree, formatter, opts, write)

    # print tree
    ret = get_tree(tree, '', sort=False, use_unicode=True)
    if ret:
        write('\n'.join(ret))


def _construct_tree(flows, tree, formatter, opts, write):
    """Construct the tree, for given flows and formatter."""
    for flow in sorted(flows, key=lambda f: f['name']):
        parts = Path(flow['name']).parts
        pointer = tree
        for part in parts[:-1]:
            if part not in pointer:
                pointer[part] = {}
            pointer = pointer[part]
        flow['name'] = parts[-1]
        item = formatter(flow, opts)
        if item is None:
            continue

        if len(parts) > 1:
            item = f' {item}'
        pointer[item] = ''


def get_pipe(opts, formatter, scan_dir=None):
    """Construct a pipe for listing flows."""
    if scan_dir:
        pipe = scan(scan_dir=scan_dir)
    elif opts.source:
        pipe = scan_multi(
            Path(path).expanduser()
            for path in glbl_cfg().get(['install', 'source dirs'])
        )
        opts.states = {'stopped'}
    else:
        pipe = scan

    show_running = 'running' in opts.states
    show_paused = 'paused' in opts.states
    show_active = show_running or show_paused or 'stopping' in opts.states
    # show_active = bool({'running', 'paused'} & opts.states)
    show_inactive = bool({'stopped'} & opts.states)

    # filter by flow name
    if opts.name:
        pipe |= filter_name(*opts.name)

    # filter by flow state
    if show_active:
        pipe |= is_active(True, filter_stop=(not show_inactive))
    elif show_inactive:
        pipe |= is_active(False)

    # get contact file information
    if show_active:
        pipe |= contact_info

    graphql_fields = {}
    graphql_filters = set()

    # filter paused/running flows
    if show_active and not (show_running and show_paused):
        graphql_fields['status'] = None
        graphql_filters.add((('status',), tuple(opts.states)))

    # get fancy data if requested
    if formatter == _format_rich:
        # graphql_fields['status'] = None
        graphql_fields.update(RICH_FIELDS)

    # add graphql queries / filters to the pipe
    if show_active and graphql_fields:
        pipe |= graphql_query(graphql_fields, filters=graphql_filters)
    elif opts.ping:
        # check the flow is running even if not required
        # by display format or filters
        pipe |= graphql_query({'status': None})

    # yield results as they are processed
    pipe.preserve_order = False

    return pipe


def get_formatter(opts):
    """Return the appropriate formatter and method for the provided opts."""
    formatter = None
    method = None

    if opts.format == 'plain':
        formatter = _format_plain
        method = _async
    elif opts.format == 'json':
        formatter = _format_json
        method = _serial
    elif opts.format == 'rich':
        formatter = _format_rich
        method = _async
    elif opts.format == 'tree':
        formatter = _format_plain
        method = _tree
    elif opts.format == 'name':  # noqa: SIM106
        formatter = _format_name_only
        method = _async
    else:
        raise NotImplementedError

    if opts.sort and method == _serial:
        raise ValueError(f'--sort is incompatible with --format {opts.format}')
    elif opts.sort:
        method = _sorted

    return formatter, method


async def scanner(opts, write, scan_dir=None):
    """Print workflows to stdout."""
    formatter, method = get_formatter(opts)
    pipe = get_pipe(opts, formatter, scan_dir)

    LOG.debug(f'pipe: {repr(pipe)}')

    await method(pipe, formatter, opts, write)


async def main(
    opts: 'Values',
    color: bool = False,
    scan_dir: Optional[Path] = None,
    write: Callable = cprint
) -> None:
    """Open up a Python API for testing purposes.

    Note:
        Don't use this API for anything other than testing, there is a
        proper Python API for these purposes.

    """
    # validate / standardise the list of workflow states
    opts.states = set(opts.states.split(','))
    if 'all' in opts.states:
        opts.states = FLOW_STATES
    elif (
        not opts.states
        or not all(
            state in FLOW_STATES
            for state in opts.states
        )
    ):
        raise InputError(
            '--states must be a comma separated list of workflow states.'
            '\nSee `cylc scan --help` for supported states.'
        )

    if not color:
        # we cannot support colour or have been requested not to use it
        opts.colour_blind = True

    # print state totals key as needed
    if opts.format == 'rich' and not opts.colour_blind:
        cprint(state_totals_key() + '\n')

    await scanner(opts, write, scan_dir)


@cli_function(get_option_parser)
def cli(_, opts: 'Values', color: bool) -> None:
    """Implement `cylc scan`."""
    asyncio.run(
        main(opts, color)
    )


ScanOptions = Options(get_option_parser())



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Cylc command argument validation logic."""


from typing import (
    Iterable,
    List,
    Optional,
)

from cylc.flow.exceptions import InputError
from cylc.flow.id import IDTokens, Tokens
from cylc.flow.task_outputs import TASK_OUTPUT_SUCCEEDED
from cylc.flow.flow_mgr import FLOW_ALL, FLOW_NEW, FLOW_NONE


ERR_OPT_FLOW_VAL = "Flow values must be an integer, or 'all', 'new', or 'none'"
ERR_OPT_FLOW_INT = "Multiple flow options must all be integer valued"
ERR_OPT_FLOW_WAIT = (
    f"--wait is not compatible with --flow={FLOW_NEW} or --flow={FLOW_NONE}"
)


def flow_opts(flows: List[str], flow_wait: bool) -> None:
    """Check validity of flow-related CLI options.

    Note the schema defaults flows to ["all"].

    Examples:
        Good:
        >>> flow_opts(["new"], False)
        >>> flow_opts(["1", "2"], False)
        >>> flow_opts(["1", "2"], True)

        Bad:
        >>> flow_opts(["none", "1"], False)
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: ... must all be integer valued

        >>> flow_opts(["cheese", "2"], True)
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: ... or 'all', 'new', or 'none'

        >>> flow_opts(["new"], True)
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: ...

    """
    for val in flows:
        val = val.strip()
        if val in [FLOW_NONE, FLOW_NEW, FLOW_ALL]:
            if len(flows) != 1:
                raise InputError(ERR_OPT_FLOW_INT)
        else:
            try:
                int(val)
            except ValueError:
                raise InputError(ERR_OPT_FLOW_VAL.format(val))

    if flow_wait and flows[0] in [FLOW_NEW, FLOW_NONE]:
        raise InputError(ERR_OPT_FLOW_WAIT)


def prereqs(prereqs: Optional[List[str]]):
    """Validate a list of prerequisites, add implicit ":succeeded".

    Comma-separated lists should be split already, client-side.

    Examples:
        # Set multiple at once:
        >>> prereqs(['1/foo:bar', '2/foo:baz'])
        ['1/foo:bar', '2/foo:baz']

        # --pre=all
        >>> prereqs(["all"])
        ['all']

        # implicit ":succeeded"
        >>> prereqs(["1/foo"])
        ['1/foo:succeeded']

        # Error: invalid format:
        >>> prereqs(["fish", "dog"])
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: ...
          * fish
          * dog

        # Error: invalid format:
        >>> prereqs(["1/foo::bar"])
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: ...
          * 1/foo::bar

        # Error: "all" must be used alone:
        >>> prereqs(["all", "2/foo:baz"])
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: --pre=all must be used alone

    """
    if prereqs is None:
        return []

    prereqs2 = []
    bad: List[str] = []
    for pre in prereqs:
        p = prereq(pre)
        if p is not None:
            prereqs2.append(p)
        else:
            bad.append(pre)
    if bad:
        raise InputError(
            "Use prerequisite format <cycle>/<task>:output\n  * "
            + "\n  * ".join(bad)
        )

    if len(prereqs2) > 1:  # noqa SIM102 (anticipates "cylc set --pre=cycle")
        if "all" in prereqs:
            raise InputError("--pre=all must be used alone")

    return prereqs2


def prereq(prereq: str) -> Optional[str]:
    """Return prereq (with :succeeded) if valid, else None.

    Format: cycle/task[:output]

    Examples:
        >>> prereq('1/foo:succeeded')
        '1/foo:succeeded'

        >>> prereq('1/foo')
        '1/foo:succeeded'

        >>> prereq('all')
        'all'

        # Error:
        >>> prereq('fish')

    """
    try:
        tokens = Tokens(prereq, relative=True)
    except ValueError:
        return None
    if (
        tokens["cycle"] == prereq
        and prereq != "all"
    ):
        # Error: --pre=<word> other than "all"
        return None

    if prereq != "all" and tokens["task_sel"] is None:
        prereq += f":{TASK_OUTPUT_SUCCEEDED}"

    return prereq


def outputs(outputs: Optional[List[str]]):
    """Validate outputs.

    Comma-separated lists should be split already, client-side.

    Examples:
        Good:
        >>> outputs(['a', 'b'])
        ['a', 'b']

        >>> outputs(["required"])  # "required" is explicit default
        []

        Bad:
        >>> outputs(["required", "a"])
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: --out=required must be used alone

        >>> outputs(["waiting"])
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: Tasks cannot be set to waiting...

    """
    # If "required" is explicit just ditch it (same as the default)
    if not outputs or outputs == ["required"]:
        return []

    if "required" in outputs:
        raise InputError("--out=required must be used alone")

    if "waiting" in outputs:
        raise InputError(
            "Tasks cannot be set to waiting. Use trigger to re-run tasks."
        )

    return outputs


def consistency(
    outputs: Optional[List[str]],
    prereqs: Optional[List[str]]
) -> None:
    """Check global option consistency

    Examples:
        >>> consistency(["a"], None)  # OK

        >>> consistency(None, ["1/a:failed"])  #OK

        >>> consistency(["a"], ["1/a:failed"])
        Traceback (most recent call last):
        cylc.flow.exceptions.InputError: ...

    """
    if outputs and prereqs:
        raise InputError("Use --prerequisite or --output, not both.")


def is_tasks(tasks: Iterable[str]):
    """All tasks in a list of tasks are task ID's without trailing job ID.

    Examples:
        # All legal
        >>> is_tasks(['1/foo', '1/bar', '*/baz', '*/*'])

        # Some legal
        >>> is_tasks(['1/foo/NN', '1/bar', '*/baz', '*/*/42'])
        Traceback (most recent call last):
        ...
        cylc.flow.exceptions.InputError: This command does not take job ids:
        * 1/foo/NN
        * */*/42

        # None legal
        >>> is_tasks(['*/baz/12'])
        Traceback (most recent call last):
        ...
        cylc.flow.exceptions.InputError: This command does not take job ids:
        * */baz/12
    """
    bad_tasks: List[str] = []
    for task in tasks:
        tokens = Tokens('//' + task)
        if tokens.lowest_token == IDTokens.Job.value:
            bad_tasks.append(task)
    if bad_tasks:
        msg = 'This command does not take job ids:\n * '
        raise InputError(msg + '\n * '.join(bad_tasks))



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Workflow event handler."""

from enum import Enum
import os
from shlex import quote
from typing import Any, Dict, List, Union, TYPE_CHECKING

from cylc.flow import LOG
from cylc.flow.cfgspec.glbl_cfg import glbl_cfg
from cylc.flow.hostuserutil import get_host, get_user
from cylc.flow.log_diagnosis import run_reftest
from cylc.flow.subprocctx import SubProcContext

if TYPE_CHECKING:
    from cylc.flow.config import WorkflowConfig
    from cylc.flow.scheduler import Scheduler


class EventData(Enum):
    """The following variables are available to workflow event handlers.

    They can be templated into event handlers with Python percent style string
    formatting e.g:

    .. code-block:: none

       %(workflow)s is running on %(host)s

    .. note::

       Substitution patterns should not be quoted in the template strings.
       This is done automatically where required.

    For an explanation of the substitution syntax, see
    `String Formatting Operations in the Python documentation
    <https://docs.python.org/3/library/stdtypes.html
    #printf-style-string-formatting>`_.

    """

    Event = 'event'
    """The type of workflow event that has occurred e.g. ``stall``."""

    Message = 'message'
    """Additional information about the event."""

    Workflow = 'workflow'
    """The workflow ID"""

    Host = 'host'
    """The host where the workflow is running."""

    Port = 'port'
    """The port where the workflow is running."""

    Owner = 'owner'
    """The user account under which the workflow is running."""

    UUID = 'uuid'
    """The unique identification string for this workflow run.

    This string is preserved for the lifetime of the scheduler and is restored
    from the database on restart.
    """

    WorkflowURL = 'workflow_url'
    """The URL defined in :cylc:conf:`flow.cylc[meta]URL`."""

    # BACK COMPAT: "suite" deprecated
    # url:
    #     https://github.com/cylc/cylc-flow/pull/4724 (& 4714)
    # from:
    #     Cylc 8
    # remove at:
    #     Cylc 8.x
    Suite = 'suite'
    """The workflow ID

    .. deprecated:: 8.0.0

       Use "workflow".
    """

    # BACK COMPAT: "suite_uuid" deprecated
    # url:
    #     https://github.com/cylc/cylc-flow/pull/4724 (& 4714)
    # from:
    #     Cylc 8
    # remove at:
    #     Cylc 8.x
    Suite_UUID = 'suite_uuid'
    """The unique identification string for this workflow run.

    .. deprecated:: 8.0.0

       Use "uuid".
    """

    # BACK COMPAT: "suite_url" deprecated
    # url:
    #     https://github.com/cylc/cylc-flow/pull/4724 (& 4714)
    # from:
    #     Cylc 8
    # remove at:
    #     Cylc 8.x
    SuiteURL = 'suite_url'
    """The URL defined in :cylc:conf:`flow.cylc[meta]URL`.

    .. deprecated:: 8.0.0

       Use "workflow_url".
    """


def construct_mail_cmd(
    subject: str, from_address: str, to_address: str
) -> List[str]:
    """Construct a mail command."""
    return [
        'mail',
        '-s', subject,
        '-r', from_address,
        to_address
    ]


def get_template_variables(
    schd: 'Scheduler',
    event: str,
    reason: str
) -> Dict[str, Union[str, int]]:
    """Return a dictionary of template varaibles for a workflow event."""
    workflow_url: str = schd.config.cfg['meta'].get('URL', '')
    return {
        # scheduler properties
        EventData.Event.value:
            event,
        EventData.Message.value:
            reason,
        EventData.Workflow.value:
            schd.workflow,
        EventData.Host.value:
            schd.host,
        EventData.Port.value:
            (schd.server.port if schd.server else -1),
        EventData.Owner.value:
            schd.owner,
        EventData.UUID.value:
            schd.uuid_str,
        EventData.WorkflowURL.value:
            workflow_url,

        # BACK COMPAT: "suite", "suite_uuid", "suite_url"
        # url:
        #     https://github.com/cylc/cylc-flow/pull/4724 (&4714)
        # from:
        #     Cylc 8
        # remove at:
        #     Cylc 8.x
        EventData.Suite.value:
            schd.workflow,
        EventData.Suite_UUID.value:
            schd.uuid_str,
        EventData.SuiteURL.value:
            workflow_url,

        # workflow metadata
        **{
            key: quote(value)
            for key, value in schd.config.cfg['meta'].items()
            if key != 'URL'
        },  # noqa: E122
    }


def process_mail_footer(
    mail_footer_tmpl: str,
    template_vars,
) -> str:
    """Process mail footer for workflow or task events.

    Returns an empty string if issues occur in processing.
    """
    try:
        return (mail_footer_tmpl + '\n') % template_vars
    except (KeyError, ValueError):
        LOG.warning(
            f'Ignoring bad mail footer template: {mail_footer_tmpl}'
        )
    return ''


class WorkflowEventHandler():
    """Workflow event handler."""

    EVENT_STARTUP = 'startup'
    EVENT_SHUTDOWN = 'shutdown'
    EVENT_ABORTED = 'abort'
    EVENT_WORKFLOW_TIMEOUT = 'workflow timeout'
    EVENT_INACTIVITY_TIMEOUT = 'inactivity timeout'
    EVENT_STALL = 'stall'
    EVENT_STALL_TIMEOUT = 'stall timeout'
    EVENT_RESTART_TIMEOUT = 'restart timeout'

    WORKFLOW_EVENT_HANDLER = 'workflow-event-handler'
    WORKFLOW_EVENT_MAIL = 'workflow-event-mail'

    def __init__(self, proc_pool):
        self.proc_pool = proc_pool

    @staticmethod
    def get_events_conf(
        config: 'WorkflowConfig', key: str, default: Any = None
    ) -> Any:
        """Return a named [scheduler][[events]] configuration."""
        for getter in (
            config.cfg['scheduler']['events'],
            config.cfg['scheduler'].get('mail', {}),
            glbl_cfg().get(['scheduler', 'events']),
            glbl_cfg().get(['scheduler', 'mail'])
        ):
            value = getter.get(key)
            if value is not None and value != []:
                return value
        return default

    def handle(self, schd: 'Scheduler', event: str, reason: str) -> None:
        """Handle a workflow event."""
        template_variables = get_template_variables(schd, event, reason)
        self._run_event_mail(schd, template_variables, event)
        self._run_event_custom_handlers(schd, template_variables, event)
        if schd.config.options.reftest and event == self.EVENT_SHUTDOWN:
            run_reftest(schd)

    def _run_event_mail(self, schd, template_variables, event):
        """Helper for "run_event_handlers", do mail notification."""
        if event in self.get_events_conf(schd.config, 'mail events', []):
            # SMTP server
            env = dict(os.environ)
            mail_smtp = self.get_events_conf(schd.config, 'smtp')
            if mail_smtp:
                env['smtp'] = mail_smtp
            subject = (
                f'[workflow %({EventData.Event.value})s]'
                f' %({EventData.Workflow.value})s' % (
                    template_variables
                )
            )
            stdin_str = ''
            for key in (
                EventData.Event.value,
                EventData.Message.value,
                EventData.Workflow.value,
                EventData.Host.value,
                EventData.Port.value,
                EventData.Owner.value,
            ):
                value = template_variables.get(key, None)
                if value:
                    stdin_str += '%s: %s\n' % (key, value)
            mail_footer_tmpl = self.get_events_conf(schd.config, 'footer')
            if mail_footer_tmpl:
                stdin_str += process_mail_footer(
                    mail_footer_tmpl,
                    template_variables,
                )
            self._send_mail(event, subject, stdin_str, schd, env)

    def _send_mail(
        self,
        event: str,
        subject: str,
        message: str,
        schd: 'Scheduler',
        env: Dict[str, str]
    ) -> None:
        proc_ctx = SubProcContext(
            (self.WORKFLOW_EVENT_HANDLER, event),
            construct_mail_cmd(
                subject,
                from_address=self.get_events_conf(
                    schd.config, 'from', f'notifications@{get_host()}'
                ),
                to_address=self.get_events_conf(schd.config, 'to', get_user())
            ),
            env=env,
            stdin_str=message
        )
        if self.proc_pool.closed:
            # Run command in foreground if process pool is closed
            self.proc_pool.run_command(proc_ctx)
            self._run_event_handlers_callback(proc_ctx)
        else:
            # Run command using process pool otherwise
            self.proc_pool.put_command(
                proc_ctx, callback=self._run_event_mail_callback)

    def _run_event_custom_handlers(self, schd, template_variables, event):
        """Helper for "run_event_handlers", custom event handlers."""
        # Look for event handlers
        # 1. Handlers for specific event
        # 2. General handlers
        config = schd.config
        handlers = self.get_events_conf(config, '%s handlers' % event)
        if not handlers and (
            event in
            self.get_events_conf(config, 'handler events', [])
        ):
            handlers = self.get_events_conf(config, 'handlers')
        if not handlers:
            return
        for i, handler in enumerate(handlers):
            cmd_key = ('%s-%02d' % (self.WORKFLOW_EVENT_HANDLER, i), event)
            try:
                cmd = handler % (template_variables)
            except KeyError as exc:
                message = f'{cmd_key} bad template: {handler}\n{exc}'
                LOG.error(message)
                continue
            if cmd == handler:
                # Nothing substituted, assume classic interface
                cmd = (
                    f"%(handler)s"
                    f" '%({EventData.Event.value})s'"
                    f" '%({EventData.Workflow.value})s'"
                    f" '%({EventData.Message.value})s'"
                ) % (
                    {'handler': handler, **template_variables}
                )
            proc_ctx = SubProcContext(
                cmd_key,
                cmd,
                env=dict(os.environ),
                shell=True  # nosec (designed to run user defined code)
            )
            if self.proc_pool.closed:
                # Run command in foreground if abort on failure is set or if
                # process pool is closed
                self.proc_pool.run_command(proc_ctx)
                self._run_event_handlers_callback(proc_ctx)
            else:
                # Run command using process pool otherwise
                self.proc_pool.put_command(
                    proc_ctx, callback=self._run_event_handlers_callback)

    @staticmethod
    def _run_event_handlers_callback(proc_ctx):
        """Callback on completion of a workflow event handler."""
        if proc_ctx.ret_code:
            msg = '%s EVENT HANDLER FAILED' % proc_ctx.cmd_key[1]
            LOG.error(str(proc_ctx))
            LOG.error(msg)
        else:
            LOG.info(str(proc_ctx))

    @staticmethod
    def _run_event_mail_callback(proc_ctx):
        """Callback the mail command for notification of a workflow event."""
        if proc_ctx.ret_code:
            LOG.warning(str(proc_ctx))
        else:
            LOG.info(str(proc_ctx))



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""cylc stop [OPTIONS] ARGS

Tell a workflow to shut down.

Examples:
  # wait for active tasks to finish, then shut down
  $ cylc stop my_workflow

  # kill active tasks, then shut down
  $ cylc stop my_workflow --kill

  # shut down immediately, leave active tasks alone
  $ cylc stop my_workflow --now

  # shut down all workflows
  $ cylc stop '*'

  # shut down after the cycle 1234 has been passed
  $ cylc stop my_workflow//1234

  # shut down after the task foo in cycle 1234 has succeeded
  $ cylc stop my_workflow//1234/foo

By default stopping workflows wait for submitted and running tasks to complete
before shutting down. You can change this behaviour with the --mode option.

There are several shutdown methods:

  1. (default) stop after current active tasks finish
  2. (--now) stop immediately, orphaning current active tasks
  3. (--kill) stop after killing current active tasks
  4. (if ID specifies a cycle point) stop after the cycle point
  5. (if ID specifies a task ID) stop after the task has succeeded
  6. (--wall-clock=T) stop after time T (an ISO 8601 date-time format e.g.
     CCYYMMDDThh:mm, CCYY-MM-DDThh, etc).

Tasks that become ready after the shutdown is ordered will be submitted
immediately if the workflow is restarted. Remaining task event handlers and
job poll and kill commands will be executed prior to shutdown, unless
--now is used twice.

This command exits immediately unless --max-polls is greater than zero, in
which case it polls to wait for workflow shutdown.
"""

from functools import partial
import sys
from typing import TYPE_CHECKING, Optional

from cylc.flow.command_polling import Poller
from cylc.flow.exceptions import (
    ClientError,
    ClientTimeout,
    CylcError,
    InputError,
    WorkflowStopped,
)
from cylc.flow.network.client_factory import get_client
from cylc.flow.network.multi import call_multi
from cylc.flow.network.schema import WorkflowStopMode
from cylc.flow.option_parsers import (
    CylcOptionParser as COP,
    ID_MULTI_ARG_DOC,
)
from cylc.flow.terminal import cli_function

if TYPE_CHECKING:
    from optparse import Values


MUTATION = '''
mutation (
  $wFlows: [WorkflowID]!,
  $stopMode: WorkflowStopMode,
  $cyclePoint: CyclePoint,
  $clockTime: TimePoint,
  $task: TaskID,
  $flowNum: Int,
) {
  stop (
    workflows: $wFlows,
    mode: $stopMode,
    cyclePoint: $cyclePoint,
    clockTime: $clockTime,
    task: $task,
    flowNum: $flowNum
  ) {
    result
  }
}
'''

POLLER_QUERY = '''
query ($wFlows: [ID]) {
  workflows(ids: $wFlows) {
    id
  }
}
'''


class StopPoller(Poller):
    """A polling object that checks if a workflow has stopped yet."""

    def __init__(self, pclient, condition, interval, max_polls):
        Poller.__init__(self, condition, interval, max_polls, None)
        self.pclient = pclient
        self.query = {
            'request_string': POLLER_QUERY,
            'variables': {'wFlows': [self.pclient.workflow]}
        }

    async def check(self):
        """Return True if workflow has stopped (success) else False"""
        try:
            await self.pclient.async_request('graphql', self.query)
        except (ClientError, ClientTimeout, CylcError):
            # failed to ping - workflow stopped or (CylcError) restarted on
            # another host:port (in which case it must have stopped first).
            return True
        else:
            # pinged - workflow must be alive
            return False


def get_option_parser() -> COP:
    parser = COP(
        __doc__,
        comms=True,
        multiworkflow=True,
        argdoc=[
            ID_MULTI_ARG_DOC,
        ],
    )

    parser.add_option(
        "-k", "--kill",
        help="Shut down after killing currently active tasks.",
        action="store_true", default=False, dest="kill")

    parser.add_option(
        "--flow", metavar="INT",
        help="Stop flow number INT from spawning more tasks. "
             "The scheduler will shut down if it is the only flow.",
        action="store", dest="flow_num")

    parser.add_option(
        "-n", "--now",
        help=(
            "Shut down without waiting for active tasks to complete." +
            " If this option is specified once," +
            " wait for task event handler, job poll/kill to complete." +
            " If this option is specified more than once," +
            " tell the workflow to terminate immediately."),
        action="count", default=0, dest="now")

    parser.add_option(
        "-w", "--wall-clock", metavar="STOP",
        help="Shut down after time STOP (ISO 8601 formatted)",
        action="store", dest="wall_clock")

    StopPoller.add_to_cmd_options(parser, d_max_polls=0)

    return parser


def _validate(
        options: 'Values',
        stop_task: Optional[str],
        stop_cycle: Optional[str],
        *globs: str
) -> None:
    """Check option choices are valid."""
    if stop_task is not None and options.kill:
        raise InputError("--kill is not compatible with stop-task")
    if stop_cycle is not None and options.kill:
        raise InputError("--kill is not compatible with stop-cycle")
    if stop_task and stop_cycle:
        raise InputError('stop-task is not compatible with stop-cycle')
    if options.kill and options.now:
        raise InputError("--kill is not compatible with --now")
    if options.flow_num and int(options.max_polls) > 0:
        raise InputError("--flow is not compatible with --max-polls")
    if options.flow_num and globs:
        raise InputError("--flow is not compatible with task IDs")


async def run(
    options: 'Values',
    workflow_id,
    *tokens_list,
) -> object:
    return await _run(options, workflow_id, *tokens_list)


async def _run(
    options: 'Values',
    workflow_id,
    *tokens_list,
) -> object:
    # parse the stop-task or stop-cycle if provided
    stop_task = stop_cycle = None
    if tokens_list:
        tokens = tokens_list[0]
        if tokens['task']:
            stop_task = tokens.relative_id
        elif tokens['cycle']:
            stop_cycle = tokens['cycle']

    _validate(options, stop_task, stop_cycle, *tokens_list)

    pclient = get_client(workflow_id, timeout=options.comms_timeout)

    if int(options.max_polls) > 0:
        # (test to avoid the "nothing to do" warning for # --max-polls=0)

        spoller = StopPoller(
            pclient,
            "workflow stopped",
            options.interval,
            options.max_polls,
        )

    # mode defaults to 'Clean'
    mode = None
    if stop_task or stop_cycle:
        pass
    elif options.kill:
        mode = WorkflowStopMode.Kill.name
    elif options.now > 1:
        mode = WorkflowStopMode.NowNow.name
    elif options.now:
        mode = WorkflowStopMode.Now.name

    mutation_kwargs = {
        'request_string': MUTATION,
        'variables': {
            'wFlows': [workflow_id],
            'stopMode': mode,
            'cyclePoint': stop_cycle,
            'clockTime': options.wall_clock,
            'task': stop_task,
            'flowNum': options.flow_num
        }
    }

    ret = await pclient.async_request('graphql', mutation_kwargs)

    if int(options.max_polls) > 0 and not await spoller.poll():
        # (test to avoid the "nothing to do" warning for # --max-polls=0)
        raise CylcError(
            f'Workflow did not shut down after {options.max_poll} polls'
        )
    return ret


@cli_function(get_option_parser)
def main(
    parser: COP,
    options: 'Values',
    *ids,
) -> None:
    rets = call_multi(
        partial(run, options),
        *ids,
        constraint='mixed',
        max_tasks=1,
        success_exceptions=(WorkflowStopped,),
    )
    sys.exit(all(rets.values()) is False)



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Manage queueing and pooling of subprocesses for the scheduler."""

from collections import deque
import json
import os
import select
from signal import SIGKILL
import sys
import shlex
from tempfile import SpooledTemporaryFile
from threading import RLock
from time import time
from subprocess import DEVNULL, run  # nosec
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Set

from cylc.flow import LOG, iter_entry_points
from cylc.flow.cfgspec.glbl_cfg import glbl_cfg
from cylc.flow.cylc_subproc import procopen
from cylc.flow.exceptions import PlatformLookupError
from cylc.flow.hostuserutil import is_remote_host
from cylc.flow.platforms import (
    log_platform_event,
    get_platform,
)
from cylc.flow.subprocctx import SubFuncContext
from cylc.flow.task_events_mgr import TaskJobLogsRetrieveContext
from cylc.flow.task_proxy import TaskProxy
from cylc.flow.wallclock import get_current_time_string

if TYPE_CHECKING:
    from subprocess import Popen
    from cylc.flow.subprocctx import SubProcContext

_XTRIG_MOD_CACHE: dict = {}
_XTRIG_FUNC_CACHE: dict = {}


def _killpg(proc, signal):
    """Kill a process group."""
    try:
        os.killpg(proc.pid, signal)
    except ProcessLookupError:
        # process group has already exited
        return False
    except PermissionError:
        # process group may contain zombie processes which will result in
        # PermissionError on some systems, not sure what happens on others
        #
        # we could go through the processes in the group and call waitpid on
        # them but waitpid is blocking and this would be a messy solution for a
        # problem that shouldn't happen (it's really a bug in the Cylc subproc)
        LOG.error(
            f'Could not kill process group: {proc.pid}'
            f'\nCommand: {" ".join(proc.args)}'
        )
        return False
    return True


def get_xtrig_mod(mod_name, src_dir):
    """Find, cache, and return a named xtrigger module.

    Locations checked in this order:
    - <src_dir>/lib/python (prepend to sys.path)
    - $CYLC_PYTHONPATH (already in sys.path)
    - `cylc.xtriggers` entry point

    (Check entry point last so users can override with local implementations).

    Workflow source dir passed in - this executes in an independent subprocess.

    Raises:
        ImportError, if the module is not found

    """
    if mod_name in _XTRIG_MOD_CACHE:
        # Found and cached already.
        return _XTRIG_MOD_CACHE[mod_name]

    # First look in <src-dir>/lib/python.
    sys.path.insert(0, os.path.join(src_dir, 'lib', 'python'))
    try:
        _XTRIG_MOD_CACHE[mod_name] = __import__(mod_name, fromlist=[mod_name])
    except ImportError:
        # Then entry point.
        for entry_point in iter_entry_points('cylc.xtriggers'):
            if mod_name == entry_point.name:
                _XTRIG_MOD_CACHE[mod_name] = entry_point.load()
                return _XTRIG_MOD_CACHE[mod_name]
        # Still unable to find anything so abort
        raise

    return _XTRIG_MOD_CACHE[mod_name]


def get_xtrig_func(mod_name, func_name, src_dir):
    """Find, cache, and return a function from an xtrigger module.

    Raises:
        ImportError, if the module is not found
        AttributeError, if the function is not found in the module

    """
    if (mod_name, func_name) in _XTRIG_FUNC_CACHE:
        return _XTRIG_FUNC_CACHE[(mod_name, func_name)]

    mod = get_xtrig_mod(mod_name, src_dir)

    _XTRIG_FUNC_CACHE[(mod_name, func_name)] = getattr(mod, func_name)

    return _XTRIG_FUNC_CACHE[(mod_name, func_name)]


def run_function(mod_name, func_name, json_args, json_kwargs, src_dir):
    """Run a Python function in the process pool.

    func_name(*func_args, **func_kwargs)

    The function is presumed to be in a module of the same name.

    Redirect any function stdout to stderr (and workflow log in debug mode).
    Return value printed to stdout as a JSON string - allows use of the
    existing process pool machinery as-is. src_dir is for local modules.

    """
    func_args = json.loads(json_args)
    func_kwargs = json.loads(json_kwargs)

    # Find and import then function.
    func = get_xtrig_func(mod_name, func_name, src_dir)

    # Redirect stdout to stderr.
    orig_stdout = sys.stdout
    sys.stdout = sys.stderr
    res = func(*func_args, **func_kwargs)

    # Restore stdout.
    sys.stdout = orig_stdout

    # Write function return value as JSON to stdout.
    sys.stdout.write(json.dumps(res))


class SubProcPool:
    """Manage queueing and pooling of subprocesses.

    Mainly used by the main loop of the scheduler, although
    the SubProcPool.run_command can be used as a standalone utility function
    to run the command in a cylc.flow.subprocctx.SubProcContext.

    A command to run under a subprocess in the pool is expected to be wrapped
    using a cylc.flow.subprocctx.SubProcContext object. The caller will add the
    context object using the SubProcPool.put_command method. A callback can
    be specified to notify the caller on exit of the subprocess.

    A command launched by the pool is expected to write to STDOUT and STDERR.
    These are captured while the command runs and/or when the command exits.
    The contents are appended to the `.out` and `.err` attributes of the
    SubProcContext object as they are read. STDIN can also be specified for the
    command. This is currently fed into the command using a temporary file.

    Note: For a cylc command that uses
    `cylc.flow.option_parsers.CylcOptionParser`, the default logging handler
    writes to the STDERR via a StreamHandler. Therefore, log messages will
    only be written to the workflow log by the callback function when the
    command exits (and only if the callback function has the logic to do so).

    """

    ERR_WORKFLOW_STOPPING = 'workflow stopping, command not run'
    JOBS_SUBMIT = 'jobs-submit'
    POLLREAD = select.POLLIN | select.POLLPRI
    RET_CODE_WORKFLOW_STOPPING = 999

    def __init__(self):
        self.size = glbl_cfg().get(['scheduler', 'process pool size'])
        self.proc_pool_timeout = glbl_cfg().get(
            ['scheduler', 'process pool timeout'])
        self.closed = False  # Close queue
        self.stopping = False  # No more job submit if True
        # .stopping may be set by an API command in a different thread
        self.stopping_lock = RLock()
        self.queuings = deque()
        self.runnings = []
        try:
            self.pipepoller = select.poll()
        except AttributeError:  # select.poll not implemented for this OS
            self.pipepoller = None

    def close(self):
        """Close pool."""
        self.set_stopping()
        self.closed = True

    @staticmethod
    def get_temporary_file():
        """Return a SpooledTemporaryFile for feeding data to command STDIN."""
        return SpooledTemporaryFile()

    def is_not_done(self):
        """Return True if queuings or runnings not empty."""
        return self.queuings or self.runnings

    def _is_stopping(self):
        """Return whether .stopping is True or not.

        Returns:
            bool: Whether the pool is stopping or not.
        """
        with self.stopping_lock:
            return self.stopping

    def _proc_exit(
        self,
        proc: 'Popen[bytes]',
        err_xtra: str,
        ctx: 'SubProcContext',
        callback: Callable,
        callback_args: list,
        bad_hosts: Optional[Set[str]] = None,
        callback_255: Optional[Callable] = None,
        callback_255_args: Optional[list] = None,
    ):
        """Get ret_code, out, err of exited command, and call its callback."""
        ctx.ret_code = proc.wait()
        out, err = (f.decode() for f in proc.communicate())
        if out:
            if ctx.out is None:
                ctx.out = ''
            ctx.out += out
        if err + err_xtra:
            if ctx.err is None:
                ctx.err = ''
            ctx.err += err + err_xtra
        LOG.debug(
            ctx.dump() if isinstance(ctx, SubFuncContext) else ctx
        )
        self._run_command_exit(
            ctx, bad_hosts=bad_hosts,
            callback=callback, callback_args=callback_args,
            callback_255=callback_255, callback_255_args=callback_255_args
        )

    def process(self):
        """Process done child processes and submit more."""
        # Handle child processes that are done
        runnings = []
        for running in self.runnings:
            (
                proc, ctx, bad_hosts,
                callback, callback_args,
                callback_255, callback_255_args
            ) = running
            # Command completed/exited
            if proc.poll() is not None:
                self._proc_exit(
                    proc, "", ctx,
                    callback=callback, callback_args=callback_args,
                    bad_hosts=bad_hosts,
                    callback_255=callback_255,
                    callback_255_args=callback_255_args
                )
                continue
            # Command timed out, kill it
            if time() > ctx.timeout:
                err_xtra = ""
                if _killpg(proc, SIGKILL):
                    err_xtra = (
                        f"\nkilled on timeout ({self.proc_pool_timeout})"
                    )
                self._proc_exit(
                    proc, err_xtra, ctx,
                    callback=callback,
                    callback_args=callback_args,
                    bad_hosts=bad_hosts
                )
                continue
            # Command still running, see if STDOUT/STDERR are readable or not
            runnings.append([
                proc, ctx, bad_hosts, callback, callback_args, None, None])
            # Unblock proc's STDOUT/STDERR if necessary. Otherwise, a full
            # STDOUT or STDERR may stop command from proceeding.
            self._poll_proc_pipes(proc, ctx)

        # Update list of running items
        self.runnings[:] = runnings
        # Create more child processes, if items in queue and space in pool
        stopping = self._is_stopping()
        while self.queuings and len(self.runnings) < self.size:
            (
                ctx, bad_hosts, callback, callback_args,
                callback_255, callback_255_args
            ) = self.queuings.popleft()
            if stopping and ctx.cmd_key == self.JOBS_SUBMIT:
                ctx.err = self.ERR_WORKFLOW_STOPPING
                ctx.ret_code = self.RET_CODE_WORKFLOW_STOPPING
                self._run_command_exit(ctx)
            else:
                proc = self._run_command_init(
                    ctx, bad_hosts, callback, callback_args,
                    callback_255, callback_255_args
                )
                if proc is not None:
                    ctx.timeout = time() + self.proc_pool_timeout
                    self.runnings.append([
                        proc, ctx, bad_hosts, callback, callback_args,
                        callback_255, callback_255_args
                    ])

    def put_command(
        self, ctx, bad_hosts=None, callback=None, callback_args=None,
        callback_255=None, callback_255_args=None
    ):
        """Queue a new shell command to execute.

        Arguments:
            ctx (cylc.flow.subprocctx.SubProcContext):
                A context object containing the command to run and its status.
            callback (callable):
                Function to call back when command exits or on error.
                Should have signature:
                    callback(ctx, *callback_args) -> None
            callback_args (list):
                Extra arguments to the callback function.
        """
        if (self.closed or self._is_stopping() and
                ctx.cmd_key == self.JOBS_SUBMIT):
            ctx.err = self.ERR_WORKFLOW_STOPPING
            ctx.ret_code = self.RET_CODE_WORKFLOW_STOPPING
            self._run_command_exit(
                ctx, bad_hosts=bad_hosts,
                callback=callback, callback_args=callback_args,
                callback_255=callback_255, callback_255_args=callback_255_args
            )
        else:
            self.queuings.append(
                [
                    ctx, bad_hosts, callback, callback_args,
                    callback_255, callback_255_args
                ]
            )

    @classmethod
    def run_command(cls, ctx):
        """Execute command in ctx and capture its output and exit status.

        Arguments:
            ctx (cylc.flow.subprocctx.SubProcContext):
                A context object containing the command to run and its status.
        """
        proc = cls._run_command_init(ctx)
        if proc:
            ctx.out, ctx.err = (f.decode() for f in proc.communicate())
            ctx.ret_code = proc.wait()
            cls._run_command_exit(ctx)

    def set_stopping(self):
        """Stop job submission."""
        with self.stopping_lock:
            self.stopping = True

    def terminate(self):
        """Drain queue, and kill and process remaining child processes."""
        self.close()
        # Drain queue
        while self.queuings:
            ctx = self.queuings.popleft()[0]
            ctx.err = self.ERR_WORKFLOW_STOPPING
            ctx.ret_code = self.RET_CODE_WORKFLOW_STOPPING
            self._run_command_exit(ctx)
        # Kill remaining processes
        for value in self.runnings:
            proc = value[0]
            if proc:
                _killpg(proc, SIGKILL)
        # Wait for child processes
        self.process()

    def _poll_proc_pipes(self, proc, ctx):
        """Poll STDOUT/ERR of proc and read some data if possible.

        This helps to unblock the command by unblocking its pipes.
        """
        if self.pipepoller is None:
            return  # select.poll not supported on this OS
        for handle in [proc.stdout, proc.stderr]:
            if not handle.closed:
                self.pipepoller.register(handle.fileno(), self.POLLREAD)
        while True:
            fileno_list = [
                fileno
                for fileno, event in self.pipepoller.poll(0.0)
                if event & self.POLLREAD]
            if not fileno_list:
                # Nothing readable
                break
            received_data = []
            for fileno in fileno_list:
                # If a file handle is readable, read something from it, add
                # results into the command context object's `.out` or `.err`,
                # whichever is relevant. To avoid blocking:
                # 1. Use `os.read` here instead of `file.read` to avoid any
                #    buffering that may cause the file handle to block.
                # 2. Call os.read only once after a poll. Poll again before
                #    another read - otherwise the os.read call may block.
                try:
                    data = os.read(fileno, 65536).decode()  # 64K
                except OSError:
                    continue
                received_data.append(data != '')
                if fileno == proc.stdout.fileno():
                    if ctx.out is None:
                        ctx.out = ''
                    ctx.out += data
                elif fileno == proc.stderr.fileno():
                    if ctx.err is None:
                        ctx.err = ''
                    ctx.err += data
            if received_data and not all(received_data):
                # if no data was pushed down the pipe exit the polling loop,
                # we can always re-enter the polling loop later if there is
                # more data
                # NOTE: this suppresses an infinite polling-loop observed
                # on darwin see:
                # https://github.com/cylc/cylc-flow/issues/3535
                # https://github.com/cylc/cylc-flow/pull/3543
                return
        self.pipepoller.unregister(proc.stdout.fileno())
        self.pipepoller.unregister(proc.stderr.fileno())

    @classmethod
    def _run_command_init(
        cls, ctx, bad_hosts=None, callback=None, callback_args=None,
        callback_255=None, callback_255_args=None
    ):
        """Prepare and launch shell command in ctx."""
        try:
            if ctx.cmd_kwargs.get('stdin_files'):
                if len(ctx.cmd_kwargs['stdin_files']) > 1:
                    stdin_file = cls.get_temporary_file()
                    for file_ in ctx.cmd_kwargs['stdin_files']:
                        if hasattr(file_, 'read'):
                            stdin_file.write(file_.read())
                        else:
                            with open(file_, 'rb') as openfile:
                                stdin_file.write(openfile.read())
                    stdin_file.seek(0)
                elif hasattr(ctx.cmd_kwargs['stdin_files'][0], 'read'):
                    stdin_file = ctx.cmd_kwargs['stdin_files'][0]
                else:
                    stdin_file = open(  # noqa: SIM115
                        # (nasty use of file handles, should avoid in future)
                        ctx.cmd_kwargs['stdin_files'][0], 'rb'
                    )
            elif ctx.cmd_kwargs.get('stdin_str'):
                stdin_file = cls.get_temporary_file()
                stdin_file.write(ctx.cmd_kwargs.get('stdin_str').encode())
                stdin_file.seek(0)
            else:
                stdin_file = DEVNULL
            proc = procopen(
                ctx.cmd, stdin=stdin_file, stdoutpipe=True, stderrpipe=True,
                # Execute command as a process group leader,
                # so we can use "os.killpg" to kill the whole group.
                preexec_fn=os.setpgrp,
                env=ctx.cmd_kwargs.get('env'),
                usesh=ctx.cmd_kwargs.get('shell'))
            # calls to open a shell are aggregated in cylc_subproc.procopen()
            # with logging for what is calling it and the commands given
        except OSError as exc:
            if exc.filename is None:
                exc.filename = ctx.cmd[0]
            LOG.exception(exc)
            ctx.ret_code = 1
            ctx.err = str(exc)
            cls._run_command_exit(
                ctx, bad_hosts=bad_hosts,
                callback=callback, callback_args=callback_args,
                callback_255=callback_255, callback_255_args=callback_255_args
            )
            return None
        else:
            LOG.debug(ctx.cmd)
            return proc

    @classmethod
    def _run_command_exit(
        cls,
        ctx: 'SubProcContext',
        bad_hosts: Optional[Set[str]] = None,
        callback: Optional[Callable] = None,
        callback_args: Optional[List[Any]] = None,
        callback_255: Optional[Callable] = None,
        callback_255_args: Optional[List[Any]] = None
    ) -> None:
        """Process command completion.

        If task has failed with a 255 error, run an alternative callback if
        one is provided.

        Args:
            ctx: SubProcContext object for this task.
            callback: Function to run on command exit.
            callback_args: Arguments to provide to callback
            callback_255: Function to run if command exits with a 255
                error - usually associated with ssh being unable to
                contact a remote host.
            callback_255_args: Arguments for the 255 callback function.

        """
        def _run_callback(callback, args_=None):
            if callable(callback):
                if not args_:
                    args_ = []
                callback(ctx, *args_)
            else:
                return False
        ctx.timestamp = get_current_time_string()

        # If cmd is fileinstall, which uses rsync, get a platform so
        # that you can use that platform's ssh command.
        platform_name = None
        platform = None
        if isinstance(ctx.cmd_key, TaskJobLogsRetrieveContext):
            try:
                platform = get_platform(ctx.cmd_key.platform_name)
            except PlatformLookupError:
                log_platform_event(
                    'Unable to retrieve job logs.',
                    {'name': ctx.cmd_key.platform_name},
                    level='warning',
                )
        elif callback_args:
            platform = callback_args[0]
            if not (
                isinstance(platform, dict)
                and 'ssh command' in platform
                and 'name' in platform
            ):
                # the first argument is not a platform
                platform = None
                # Backup, get a platform name from the config:
                for arg in callback_args:
                    if isinstance(arg, TaskProxy):
                        platform_name = arg.tdef.rtconfig['platform']
                    elif (
                        isinstance(arg, list)
                        and isinstance(arg[0], TaskProxy)
                    ):
                        platform_name = arg[0].tdef.rtconfig['platform']

        if cls.ssh_255_fail(ctx) or cls.rsync_255_fail(ctx, platform) is True:
            # Job log retrieval passes a special object as a command key
            # Extra logic to provide sensible strings for logging.
            if isinstance(ctx.cmd_key, TaskJobLogsRetrieveContext):
                cmd_key = ctx.cmd_key.key
            else:
                cmd_key = ctx.cmd_key
            log_platform_event(
                # NOTE: the failure of the command should be logged elsewhere
                (
                    f'Could not connect to {ctx.host}.'
                    f'\n* {ctx.host} has been added to the list of'
                    ' unreachable hosts'
                    f'\n* {cmd_key} will retry if another host is available.'
                ),
                platform or {'name': platform_name},
                level='warning',
            )

            # If callback_255 takes the same args as callback, we don't
            # want to spec those args:
            if callable(callback_255) and callback_255_args is None:
                callback_255_args = callback_args

            # Run Callback
            if bad_hosts is not None:
                bad_hosts.add(ctx.host)

            res = _run_callback(callback_255, callback_255_args)
            if res is False:
                _run_callback(callback, callback_args)
        else:
            # For every other return code run default callback.
            _run_callback(callback, callback_args)

    @staticmethod
    def ssh_255_fail(ctx) -> bool:
        """Test context for ssh command failing with a 255 error."""
        ssh_255_fail = False
        if (
            ctx.cmd[0] == 'ssh'
            and ctx.ret_code == 255
        ):
            ssh_255_fail = True
        return ssh_255_fail

    @staticmethod
    def rsync_255_fail(ctx, platform=None) -> bool:
        """Tests context for rsync failing to communicate with a host.

        If there has been a failure caused by rsync being unable to connect
        try a test of ssh connectivity. Necessary because loss of connectivity
        may cause different rsync failures depending on version, and some of
        the failures may be caused by other problems.
        """
        rsync_255_fail = False
        platform_rsync_cmd = (
            platform['rsync command']
            if platform is not None
            else 'rsync'
        )
        rsync_cmd = shlex.split(platform_rsync_cmd)
        if (
            ctx.cmd[0] == rsync_cmd[0]
            and ctx.ret_code not in [0, 255]
            and is_remote_host(ctx.host)
        ):
            ssh_cmd = (
                platform['ssh command']
                if platform is not None
                else 'ssh'
            )
            ssh_test_cmd = shlex.split(f'{ssh_cmd} {ctx.host} true')
            LOG.info(
                f'testing connectivity for {ctx.host} using {ssh_test_cmd}'
            )
            ssh_test = run(
                shlex.split(f'{ssh_cmd} {ctx.host} true'),  # nosec B603 *
                capture_output=True
            )
            # * (command is trusted input)
            if ssh_test.returncode == 255:
                rsync_255_fail = True
        elif (
            ctx.cmd[0] == rsync_cmd[0]
            and ctx.ret_code == 255
        ):
            rsync_255_fail = True
        return rsync_255_fail



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

r"""cylc workflow-state [OPTIONS] ARGS

Check or poll a workflow database for task statuses or completed outputs.

The ID argument can target a workflow, or a cycle point, or a specific
task, with an optional selector on cycle or task to match task status,
output trigger (if not a status, or with --trigger) or output message
(with --message). All matching results will be printed.

If no results match, the command will repeatedly check (poll) until a match
is found or polling is exhausted (see --max-polls and --interval). For a
one-off check set --max-polls=1.

If the database does not exist at first, polls are consumed waiting for it
so you can start checking before the target workflow is started.

Legacy (pre-8.3.0) options are supported, but deprecated, for existing scripts:
    cylc workflow-state --task=NAME --point=CYCLE --status=STATUS
        --output=MESSAGE --message=MESSAGE --task-point WORKFLOW
(Note from 8.0 until 8.3.0 --output and --message both match task messages).

In "cycle/task:selector" the selector will match task statuses, unless:
  - if it is not a known status, it will match task output triggers
    (Cylc 8 DB) or task ouput messages (Cylc 7 DB)
  - with --triggers, it will only match task output triggers
  - with --messages (deprecated), it will only match task output messages.
    Triggers are more robust - they match manually and naturally set outputs.

Selector does not default to "succeeded". If omitted, any status will match.

The "finished" pseudo-output is an alias for "succeeded or failed".

In the ID, both cycle and task can include "*" to match any sequence of zero
or more characters. Quote the pattern to protect it from shell expansion.

Note tasks get recorded in the DB once they enter the active window (n=0).

Flow numbers are only printed for flow numbers > 1.

USE IN TASK SCRIPTING:
  - To poll a task at the same cycle point in another workflow, just use
    $CYLC_TASK_CYCLE_POINT in the ID.
  - To poll a task at an offset cycle point, use the --offset option to
    have Cylc do the datetime arithmetic for you.
  - However, see also the workflow_state xtrigger for this use case.

WARNINGS:
 - Typos in the workflow or task ID will result in fruitless polling.
 - To avoid missing transient states ("submitted", "running") poll for the
   corresponding output trigger instead ("submitted", "started").
 - Cycle points are auto-converted to the DB point format (and UTC mode).
 - Task outputs manually completed by "cylc set" have "(force-completed)"
   recorded as the task message in the DB, so it is best to query trigger
   names, not messages, unless specifically interested in forced outputs.

Examples:

  # Print the status of all tasks in WORKFLOW:
  $ cylc workflow-state WORKFLOW

  # Print the status of all tasks in cycle point 2033:
  $ cylc workflow-state WORKFLOW//2033

  # Print the status of all tasks named foo:
  $ cylc workflow-state "WORKFLOW//*/foo"

  # Print all succeeded tasks:
  $ cylc workflow-state "WORKFLOW//*/*:succeeded"

  # Print all tasks foo that completed output (trigger name) file1:
  $ cylc workflow-state "WORKFLOW//*/foo:file1"

  # Print if task 2033/foo completed output (trigger name) file1:
  $ cylc workflow-state WORKFLOW//2033/foo:file1

See also:
  - the workflow_state xtrigger, for state polling within workflows
  - "cylc dump -t", to query a scheduler for current statuses
  - "cylc show", to query a scheduler for task prerequisites and outputs
"""

import asyncio
import os
import sqlite3
import sys
from typing import TYPE_CHECKING, List, Optional

from cylc.flow.pathutil import get_cylc_run_dir
from cylc.flow.id import Tokens
from cylc.flow.exceptions import InputError
from cylc.flow.option_parsers import (
    ID_SEL_ARG_DOC,
    CylcOptionParser as COP,
)
from cylc.flow import LOG
from cylc.flow.command_polling import Poller
from cylc.flow.dbstatecheck import CylcWorkflowDBChecker
from cylc.flow.terminal import cli_function
from cylc.flow.workflow_files import infer_latest_run_from_id
from cylc.flow.task_state import TASK_STATUSES_ORDERED

if TYPE_CHECKING:
    from optparse import Values


WILDCARD = "*"

# polling defaults
MAX_POLLS = 12
INTERVAL = 5

OPT_DEPR_MSG = "DEPRECATED, use ID"
OPT_DEPR_MSG1 = 'DEPRECATED, use "ID:STATUS"'
OPT_DEPR_MSG2 = 'DEPRECATED, use "ID:MSG"'


def unquote(s: str) -> str:
    """Remove leading & trailing quotes from a string.

    Examples:
    >>> unquote('"foo"')
    'foo'
    >>> unquote("'foo'")
    'foo'
    >>> unquote('foo')
    'foo'
    >>> unquote("'tis a fine morning")
    "'tis a fine morning"
    """
    if (
        s.startswith('"') and s.endswith('"')
        or s.startswith("'") and s.endswith("'")
    ):
        return s[1:-1]
    return s


class WorkflowPoller(Poller):
    """An object that polls for task states or outputs in a workflow DB."""

    def __init__(
        self,
        id_: str,
        offset: Optional[str],
        flow_num: Optional[int],
        alt_cylc_run_dir: Optional[str],
        default_status: Optional[str],
        is_trigger: bool,
        is_message: bool,
        old_format: bool = False,
        pretty_print: bool = False,
        **kwargs
    ):
        self.id_ = id_
        self.offset = offset
        self.flow_num = flow_num
        self.alt_cylc_run_dir = alt_cylc_run_dir
        self.old_format = old_format
        self.pretty_print = pretty_print

        try:
            tokens = Tokens(self.id_)
        except ValueError as exc:
            raise InputError(exc)

        self.workflow_id_raw = tokens.workflow_id
        self.selector = (
            tokens["cycle_sel"] or
            tokens["task_sel"] or
            default_status
        )
        if self.selector:
            self.selector = unquote(self.selector)
        self.cycle_raw = tokens["cycle"]
        self.task = tokens["task"]

        self.workflow_id: Optional[str] = None
        self.cycle: Optional[str] = None
        self.result: Optional[List[List[str]]] = None
        self._db_checker: Optional[CylcWorkflowDBChecker] = None

        self.is_message = is_message
        if is_message:
            self.is_trigger = False
        else:
            self.is_trigger = (
                is_trigger or
                (
                    self.selector is not None and
                    self.selector not in TASK_STATUSES_ORDERED
                )
            )
        super().__init__(**kwargs)

    def _find_workflow(self) -> bool:
        """Find workflow and infer run directory, return True if found."""
        try:
            self.workflow_id = infer_latest_run_from_id(
                self.workflow_id_raw,
                self.alt_cylc_run_dir
            )
        except InputError:
            LOG.debug("Workflow not found")
            return False

        if self.workflow_id != self.workflow_id_raw:
            # Print inferred ID.
            sys.stderr.write(f"Inferred workflow ID: {self.workflow_id}\n")
        return True

    @property
    def db_checker(self) -> Optional[CylcWorkflowDBChecker]:
        """Connect to workflow DB if not already connected.

        Returns DB checker if connected.
        """
        if not self._db_checker:
            try:
                self._db_checker = CylcWorkflowDBChecker(
                    get_cylc_run_dir(self.alt_cylc_run_dir),
                    self.workflow_id
                )
            except (OSError, sqlite3.Error):
                LOG.debug("DB not connected")
                return None

        return self._db_checker

    async def check(self) -> bool:
        """Return True if requested state achieved, else False.

        Called once per poll by super() so only find and connect once.

        Store self.result for external access.

        """
        if self.workflow_id is None and not self._find_workflow():
            return False

        if self.db_checker is None:
            return False

        if self.cycle is None:
            # Adjust target cycle point to the DB format.
            self.cycle = self.db_checker.adjust_point_to_db(
                self.cycle_raw, self.offset)

        self.result = self.db_checker.workflow_state_query(
            self.task, self.cycle, self.selector, self.is_trigger,
            self.is_message, self.flow_num
        )
        if self.result:
            # End the polling dot stream and print inferred runN workflow ID.
            self.db_checker.display_maps(
                self.result, self.old_format, self.pretty_print)

        return bool(self.result)


def get_option_parser() -> COP:
    parser = COP(
        __doc__,
        argdoc=[ID_SEL_ARG_DOC]
    )

    # --run-dir for pre-8.3.0 back-compat
    parser.add_option(
        "-d", "--alt-cylc-run-dir", "--run-dir",
        help="Alternate cylc-run directory, e.g. for other users' workflows.",
        metavar="DIR", action="store", dest="alt_cylc_run_dir", default=None)

    parser.add_option(
        "-s", "--offset",
        help="Offset from ID cycle point as an ISO8601 duration for datetime"
        " cycling (e.g. 'PT30M' for 30 minutes) or an integer interval for"
        " integer cycling (e.g. 'P2'). This can be used in task job scripts"
        " to poll offset cycle points without doing the cycle arithmetic"
        " yourself - but see also the workflow_state xtrigger.",
        action="store", dest="offset", metavar="DURATION", default=None)

    parser.add_option(
        "--flow",
        help="Flow number, for target tasks. By default, any flow.",
        action="store", type="int", dest="flow_num", default=None)

    parser.add_option(
        "--triggers",
        help="Task selector should match output triggers rather than status."
             " (Note this is not needed for custom outputs).",
        action="store_true", dest="is_trigger", default=False)

    parser.add_option(
        "--messages",
        help="Task selector should match output messages rather than status.",
        action="store_true", dest="is_message", default=False)

    parser.add_option(
        "--pretty",
        help="Pretty-print outputs (the default is single-line output).",
        action="store_true", dest="pretty_print", default=False)

    parser.add_option(
        "--old-format",
        help="Print results in legacy comma-separated format.",
        action="store_true", dest="old_format", default=False)

    # Back-compat support for pre-8.3.0 command line options.
    parser.add_option(
        "-t", "--task", help=f"Task name. {OPT_DEPR_MSG}.",
        metavar="NAME",
        action="store", dest="depr_task", default=None)

    parser.add_option(
        "-p", "--point", metavar="CYCLE",
        help=f"Cycle point. {OPT_DEPR_MSG}.",
        action="store", dest="depr_point", default=None)

    parser.add_option(
        "-T", "--task-point",
        help="Get cycle point from the environment variable"
        " $CYLC_TASK_CYCLE_POINT (e.g. in task job scripts)",
        action="store_true", dest="depr_env_point", default=False)

    parser.add_option(
        "-S", "--status",
        metavar="STATUS",
        help=f"Task status. {OPT_DEPR_MSG1}.",
        action="store", dest="depr_status", default=None)

    # Prior to 8.3.0 --output was just an alias for --message
    parser.add_option(
        "-O", "--output", "-m", "--message",
        metavar="MSG",
        help=f"Task output message. {OPT_DEPR_MSG2}.",
        action="store", dest="depr_msg", default=None)

    WorkflowPoller.add_to_cmd_options(
        parser,
        d_interval=INTERVAL,
        d_max_polls=MAX_POLLS
    )

    return parser


@cli_function(get_option_parser, remove_opts=["--db"])
def main(parser: COP, options: 'Values', *ids: str) -> None:

    # Note it would be cleaner to use 'id_cli.parse_ids()' here to get the
    # workflow ID and tokens, but that function infers run number and fails
    # if the workflow is not installed yet. We want to be able to start polling
    # before the workflow is installed, which makes it easier to get a set of
    # interdependent workflows up and running, so runN inference is done inside
    # the poller. TODO: consider using id_cli.parse_ids inside the poller.
    # (Note this applies to polling tasks, which use the CLI, not xtriggers).

    id_ = ids[0].rstrip('/')  # might get 'id/' due to autcomplete

    if any(
        [
            options.depr_task,
            options.depr_status,
            options.depr_msg,  # --message and --trigger
            options.depr_point,
            options.depr_env_point
        ]
    ):
        depr_opts = (
            "--task, --status, --message, --output, --point, --task-point"
        )

        if id_ != Tokens(id_)["workflow"]:
            raise InputError(
                f"with deprecated {depr_opts}, the argument must be a"
                " plain workflow ID (i.e. no cycle, task, or :selector)."
            )

        if options.depr_status and options.depr_msg:
            raise InputError("set --status or --message, not both.")

        if options.depr_env_point:
            if options.depr_point:
                raise InputError(
                    "set --task-point or --point=CYCLE, not both.")
            try:
                options.depr_point = os.environ["CYLC_TASK_CYCLE_POINT"]
            except KeyError:
                raise InputError(
                    "--task-point: $CYLC_TASK_CYCLE_POINT is not defined")

        if options.depr_point is not None:
            id_ += f"//{options.depr_point}"
        elif (
            options.depr_task is not None or
            options.depr_status is not None or
            options.depr_msg is not None
        ):
            id_ += "//*"
        if options.depr_task is not None:
            id_ += f"/{options.depr_task}"
        if options.depr_status is not None:
            id_ += f":{options.depr_status}"
        elif options.depr_msg is not None:
            id_ += f":{options.depr_msg}"
            options.is_message = True

        msg = f"{depr_opts} are deprecated. Please use an ID: "
        if not options.depr_env_point:
            msg += id_
        else:
            msg += id_.replace(options.depr_point, "$CYLC_TASK_CYCLE_POINT")
        LOG.warning(msg)

    poller = WorkflowPoller(
        id_,
        options.offset,
        options.flow_num,
        options.alt_cylc_run_dir,
        default_status=None,
        is_trigger=options.is_trigger,
        is_message=options.is_message,
        old_format=options.old_format,
        pretty_print=options.pretty_print,
        condition=id_,
        interval=options.interval,
        max_polls=options.max_polls,
        args=None
    )

    if not asyncio.run(
        poller.poll()
    ):
        sys.exit(1)



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions to return paths to common workflow files and directories."""

import errno
import os
from pathlib import Path
import re
from shutil import rmtree
from time import sleep
from typing import Dict, Iterable, Set, Union, Optional, Any

from cylc.flow import LOG
from cylc.flow.cfgspec.glbl_cfg import glbl_cfg
from cylc.flow.exceptions import (
    FileRemovalError, InputError, WorkflowFilesError
)
from cylc.flow.platforms import get_localhost_install_target


# Note: do not import this elsewhere, as it might bypass unit test
# monkeypatching:
_CYLC_RUN_DIR = os.path.join('$HOME', 'cylc-run')

EXPLICIT_RELATIVE_PATH_REGEX = re.compile(
    rf'''
    ^({re.escape(os.curdir)}|{re.escape(os.pardir)})
    ({re.escape(os.sep)}|$)
    ''',
    re.VERBOSE
)
"""Matches relative paths that are explicit (starts with ./)"""

SHELL_ENV_VARS = re.compile(r'\$[^$/]*')


def expand_path(*args: Union[Path, str]) -> str:
    """Expand both vars and user in path and normalise it, joining any
    extra args."""
    return os.path.normpath(os.path.expanduser(os.path.expandvars(
        os.path.join(*args)
    )))


def get_remote_workflow_run_dir(
    workflow_id: Union[Path, str], *args: Union[Path, str]
) -> str:
    """Return remote workflow run directory, joining any extra args,
    NOT expanding vars or user."""
    return os.path.join(_CYLC_RUN_DIR, workflow_id, *args)


def get_remote_workflow_run_job_dir(
    workflow_id: Union[Path, str], *args: Union[Path, str]
) -> str:
    """Return remote workflow job log directory, joining any extra args,
    NOT expanding vars or user."""
    return get_remote_workflow_run_dir(workflow_id, 'log', 'job', *args)


def get_cylc_run_dir(alt_run_dir: Optional[str] = None) -> str:
    """Return the cylc-run dir, or alt path, with vars/user expanded."""
    return expand_path(alt_run_dir or _CYLC_RUN_DIR)


def get_alt_workflow_run_dir(
    alt_run_dir: Union[Path, str],
    workflow_id: Union[Path, str],
    *args: Union[Path, str]
) -> str:
    """Return alternate workflow run directory.

    Join any extra args, and expand vars and user.
    Does not check that the directory exists.
    """
    return expand_path(alt_run_dir, workflow_id, *args)


def get_workflow_run_dir(
    workflow_id: Union[Path, str], *args: Union[Path, str]
) -> str:
    """Return local workflow run directory.

    Join any extra args, and expand vars and user.
    Does not check that the directory exists.
    """
    return expand_path(_CYLC_RUN_DIR, workflow_id, *args)


def get_workflow_run_job_dir(workflow, *args):
    """Return workflow run job (log) directory, join any extra args."""
    return get_workflow_run_dir(workflow, 'log', 'job', *args)


def get_workflow_run_scheduler_log_dir(workflow, *args):
    """Return workflow run log directory, join any extra args."""
    return get_workflow_run_dir(workflow, 'log', 'scheduler', *args)


def get_workflow_run_scheduler_log_path(workflow):
    """Return workflow run log file path."""
    return get_workflow_run_scheduler_log_dir(workflow, 'log')


def get_workflow_file_install_log_dir(workflow, *args):
    """Return workflow file install log file dir, join any extra args."""
    return get_workflow_run_dir(
        workflow, 'log', 'remote-install', *args
    )


def get_workflow_run_config_log_dir(workflow, *args):
    """Return workflow run flow.cylc log directory, join any extra args."""
    return get_workflow_run_dir(workflow, 'log', 'config', *args)


def get_workflow_run_pub_db_path(workflow):
    """Return workflow run public database file path."""
    return get_workflow_run_dir(workflow, 'log', 'db')


def get_workflow_run_share_dir(workflow, *args):
    """Return local workflow work/share directory, join any extra args."""
    return get_workflow_run_dir(workflow, 'share', *args)


def get_workflow_run_work_dir(workflow, *args):
    """Return local workflow work/work directory, join any extra args."""
    return get_workflow_run_dir(workflow, 'work', *args)


def get_workflow_test_log_path(workflow):
    """Return workflow run ref test log file path."""
    return get_workflow_run_scheduler_log_dir(workflow, 'reftest.log')


def make_workflow_run_tree(workflow):
    """Create all top-level cylc-run output dirs on the workflow host."""
    for dir_ in (
        get_workflow_run_dir(workflow),
        get_workflow_run_scheduler_log_dir(workflow),
        get_workflow_run_job_dir(workflow),
        get_workflow_run_config_log_dir(workflow),
        get_workflow_run_share_dir(workflow),
        get_workflow_run_work_dir(workflow),
    ):
        if not Path(dir_).is_dir():
            os.makedirs(dir_)
            LOG.debug(f'{dir_}: directory created')


def make_localhost_symlinks(
    rund: Union[Path, str],
    named_sub_dir: str,
    symlink_conf: Optional[Dict[str, Dict[str, str]]] = None
) -> Dict[str, Union[Path, str]]:
    """Creates symlinks for any configured symlink dirs from glbl_cfg.
    Args:
        rund: the entire run directory path
        named_sub_dir: e.g workflow_name/run1
        symlink_conf: Symlinks dirs configuration passed from cli

    Returns:
        Dictionary of symlinks with sources as keys and
        destinations as values: ``{target: symlink}``

    """
    symlinks_created = {}
    dirs_to_symlink = get_dirs_to_symlink(
        get_localhost_install_target(),
        named_sub_dir, symlink_conf=symlink_conf
    )
    for key, value in dirs_to_symlink.items():
        if value is None:
            continue
        if key == 'run':
            symlink_path = rund
        else:
            symlink_path = os.path.join(rund, key)
        target = expand_path(value)
        env_vars = SHELL_ENV_VARS.findall(target)
        if env_vars:
            raise WorkflowFilesError(
                f"Can't symlink to {target}\n"
                "Undefined variables, check "
                f"global config: {', '.join(env_vars)}")

        symlink_success = make_symlink_dir(symlink_path, target)
        # Symlink info returned for logging purposes. Symlinks should be
        # created before logs as the log dir may be a symlink.
        if symlink_success:
            symlinks_created[target] = symlink_path
    return symlinks_created


def get_dirs_to_symlink(
    install_target: str,
    workflow_id: str,
    symlink_conf: Optional[Dict[str, Dict[str, Any]]] = None
) -> Dict[str, str]:
    """Returns dictionary of directories to symlink.

    Note the paths should remain unexpanded, to be expanded on the remote.

    Args:
        install_target: Symlinks to be created on this install target
        workflow_id: full id of the run, e.g. myflow/run1
        symlink_conf: Symlink dirs, if sent on the cli.
            Defaults to None, in which case global config symlink dirs will
            be applied.

    Returns:
        dirs_to_symlink: [directory: symlink_path]
    """
    dirs_to_symlink: Dict[str, str] = {}
    if symlink_conf is None:
        symlink_conf = glbl_cfg().get(['install', 'symlink dirs'])
    if install_target not in symlink_conf.keys():
        return dirs_to_symlink
    base_dir = symlink_conf[install_target]['run']
    if base_dir:
        dirs_to_symlink['run'] = os.path.join(
            base_dir, 'cylc-run', workflow_id)
    for dir_ in ['log', 'share', 'share/cycle', 'work']:
        link = symlink_conf[install_target].get(dir_, None)
        if (not link) or link == base_dir:
            continue
        dirs_to_symlink[dir_] = os.path.join(
            link, 'cylc-run', workflow_id, dir_)
    return dirs_to_symlink


def make_symlink_dir(path: Union[Path, str], target: Union[Path, str]) -> bool:
    """Makes symlinks for directories.

    Args:
        path: Absolute path of the desired symlink.
        target: Absolute path of the symlink's target directory.

    Returns True if symlink created, or False if skipped.
    """
    path = Path(path)
    target = Path(target)
    if path.exists():
        # note all three checks are needed here due to case where user has set
        # their own symlink which does not match the global config set one.
        if path.is_symlink() and target.exists() and path.samefile(target):
            # correct symlink already exists
            return False
        # symlink name is in use by a physical file or directory
        LOG.warning(
            f"Path {path} already exists. Cannot create symlink to {target}."
        )
        return False
    elif path.is_symlink():
        # remove a bad symlink.
        try:
            path.unlink()
        except OSError as exc:
            raise WorkflowFilesError(
                f"Failed to remove broken symlink {path}\n{exc}"
            )
    try:
        target.mkdir(parents=True, exist_ok=False)
    except FileExistsError:
        raise WorkflowFilesError(
            f"Symlink dir target already exists: ({path} ->) {target}\n"
            "Tip: in future, use 'cylc clean' instead of manually deleting "
            "workflow run dirs."
        )

    # This is needed in case share and share/cycle have the same symlink dir:
    if path.exists():
        return False

    path.parent.mkdir(parents=True, exist_ok=True)
    try:
        path.symlink_to(target)
        return True
    except OSError as exc:
        raise WorkflowFilesError(f"Error when symlinking\n{exc}")


def remove_dir_and_target(path: Union[Path, str]) -> None:
    """Delete a directory tree (i.e. including contents), as well as the
    target directory tree if the specified path is a symlink.

    Args:
        path: the absolute path of the directory to delete.
    """
    if not os.path.isabs(path):
        raise ValueError('Path must be absolute')
    if os.path.exists(path) and not os.path.isdir(path):
        raise NotADirectoryError(path)
    if os.path.islink(path):
        if os.path.exists(path):
            target = os.path.realpath(path)
            LOG.info(
                "Removing symlink and its target directory: "
                f"{path} -> {target}"
            )
            _rmtree(target)
        else:
            LOG.info(f'Removing broken symlink: {path}')
        os.remove(path)
    elif not os.path.exists(path):
        raise FileNotFoundError(path)
    else:
        LOG.info(f'Removing directory: {path}')
        _rmtree(path)


def _rmtree(
    target: Union[Path, str],
    retries: int = 10,
    sleep_time: float = 1,
):
    """Make rmtree more robust to nfs issues.

    If a file is deleted which is being held open for reading by
    another process. NFS will create a ".nfs" file in the
    containing directory to handle this.

    If you try to delete the directory which contains these
    files you will get either a ENOTEMPTY or EBUSY error.

    A likely cause of open file handles in cylc-run directories
    is `cylc cat-log -m t`. If the file being cat-log'ged is removed,
    the command will fail on its next poll. The default poll
    interval is one second, so if we wait a couple of seconds and
    retry the removal it will likely work.

    This command retries removal a specified number
    of times at a specified interval before failing to
    give cat-log process a chance to die gracefully and
    release their filesystem locks. For more info see:
    https://github.com/cylc/cylc-flow/pull/5359#issuecomment-1479989975
    """
    for _try_num in range(retries):
        try:
            rmtree(target)
            return
        except OSError as exc:
            if exc.errno in {errno.ENOTEMPTY, errno.EBUSY}:
                err = exc
                sleep(sleep_time)
            else:
                raise
    raise FileRemovalError(err)


def remove_dir_or_file(path: Union[Path, str]) -> None:
    """Delete a directory tree, or a file, or a symlink.
    Does not follow symlinks.

    Args:
        path: the absolute path of the directory/file/symlink to delete.
    """
    if not os.path.isabs(path):
        raise ValueError("Path must be absolute")
    if os.path.islink(path):
        LOG.info(f"Removing symlink: {path}")
        os.remove(path)
    elif os.path.isfile(path):
        LOG.info(f"Removing file: {path}")
        os.remove(path)
    else:
        LOG.info(f"Removing directory: {path}")
        _rmtree(path)


def remove_empty_parents(
    path: Union[Path, str], tail: Union[Path, str]
) -> None:
    """Work our way up the tail of path, removing empty dirs only.

    Args:
        path: Absolute path to the directory, e.g. /foo/bar/a/b/c
        tail: The tail of the path to work our way up, e.g. a/b/c

    Example:
        remove_empty_parents('/foo/bar/a/b/c', 'a/b/c') would remove
        /foo/bar/a/b (assuming it's empty), then /foo/bar/a (assuming it's
        empty).
    """
    path = Path(path)
    if not path.is_absolute():
        raise ValueError('path must be absolute')
    tail = Path(tail)
    if tail.is_absolute():
        raise ValueError('tail must not be an absolute path')
    if not str(path).endswith(str(tail)):
        raise ValueError(f"path '{path}' does not end with '{tail}'")
    depth = len(tail.parts) - 1
    for i in range(depth):
        parent = path.parents[i]
        if not parent.is_dir():
            continue
        try:
            parent.rmdir()
            LOG.info(f'Removing directory: {parent}')
        except OSError:
            break


def get_next_rundir_number(run_path: Union[str, Path]) -> int:
    """Return the next run number for a new install.

    Args:
        run_path: Top level installed workflow dir
        (often ``~/cylc-run/workflow``).

    """
    re_runX = re.compile(r'run(\d+)$')
    run_n_path = Path(os.path.expanduser(os.path.join(run_path, "runN")))
    if run_n_path.exists() and run_n_path.is_symlink():
        old_run_path = os.readlink(run_n_path)
        # Line below could in theory not return a match group, so mypy objects.
        # This function unlikely to be called in circumstances where this will
        # be a problem.
        last_run_num = re_runX.search(  # type: ignore
            old_run_path).group(1)
        last_run_num = int(last_run_num)
    else:
        # If the ``runN`` symlink has been removed, get next numbered run from
        # file names:
        paths = Path(run_path).glob('run[0-9]*')
        run_numbers = (
            int(m.group(1)) for m in (
                re_runX.search(i.name) for i in paths
            ) if m
        )
        last_run_num = max(run_numbers, default=0)

    return last_run_num + 1


def parse_rm_dirs(rm_dirs: Iterable[str]) -> Set[str]:
    """Parse a list of possibly colon-separated dirs (or files or globs).
    Return the set of all the dirs.

    Used by cylc clean with the --rm option.
    """
    result: Set[str] = set()
    for item in rm_dirs:
        for part in item.split(':'):
            part = part.strip()
            if not part:
                continue
            is_dir = part.endswith(os.sep)
            part = os.path.normpath(part)
            if os.path.isabs(part):
                raise InputError("--rm option cannot take absolute paths")
            if (
                part in {os.curdir, os.pardir} or
                part.startswith(f"{os.pardir}{os.sep}")  # '../'
            ):
                raise InputError(
                    "--rm option cannot take paths that point to the "
                    "run directory or above"
                )
            if is_dir:
                # Preserve trailing slash to ensure it only matches dirs,
                # not files, when globbing
                part += os.sep
            result.add(part)
    return result


def is_relative_to(path1: Union[Path, str], path2: Union[Path, str]) -> bool:
    """Return whether or not path1 is relative to path2.

    Normalizes both paths to avoid trickery with paths containing `..`
    somewhere in them.
    """
    # In future, we can just use pathlib.Path.is_relative_to()
    # when Python 3.9 becomes the minimum supported version
    try:
        Path(os.path.normpath(path1)).relative_to(os.path.normpath(path2))
    except ValueError:
        return False
    return True


def get_workflow_name_from_id(workflow_id: str) -> str:
    """Workflow name is the ID shorn of the runN directory name.
    """
    cylc_run_dir = Path(get_cylc_run_dir())
    if Path(workflow_id).is_absolute():
        # this is a source directory, not an install dir:
        return workflow_id
    else:
        id_path = cylc_run_dir / workflow_id
    name_path = id_path

    # Look for ``id_path.parent/_cylc_install`` first because expected to
    # be most common:
    if (id_path.parent / '_cylc-install').is_dir():
        name_path = Path(id_path).parent
    elif (id_path / '_cylc-install').is_dir():
        name_path = id_path

    return str(name_path.relative_to(cylc_run_dir))



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Provide a class to represent a task proxy in a running workflow."""

from collections import Counter
from copy import copy
from fnmatch import fnmatchcase
from time import time
from typing import (
    Any,
    Callable,
    Counter as TypingCounter,
    Dict,
    List,
    Optional,
    Set,
    TYPE_CHECKING,
    Tuple,
)

from metomi.isodatetime.timezone import get_local_time_zone

from cylc.flow import LOG
from cylc.flow.flow_mgr import stringify_flow_nums
from cylc.flow.platforms import get_platform
from cylc.flow.task_action_timer import TimerFlags
from cylc.flow.task_state import (
    TaskState,
    TASK_STATUS_WAITING,
    TASK_STATUS_EXPIRED,
)
from cylc.flow.taskdef import generate_graph_children
from cylc.flow.wallclock import get_unix_time_from_time_string as str2time
from cylc.flow.cycling.iso8601 import (
    point_parse,
    interval_parse,
    ISO8601Interval
)

if TYPE_CHECKING:
    from cylc.flow.cycling import PointBase
    from cylc.flow.simulation import ModeSettings
    from cylc.flow.task_action_timer import TaskActionTimer
    from cylc.flow.taskdef import TaskDef
    from cylc.flow.id import Tokens


class TaskProxy:
    """Represent an instance of a cycling task in a running workflow.

    Attributes:
        .clock_trigger_times:
            Memoization of clock trigger times (Used for wall_clock xtrigger):
            {offset string: seconds from epoch}
        .expire_time:
            Time in seconds since epoch when this task is considered expired.
        .identity:
            Task ID in POINT/NAME syntax.
        .tokens:
            Task ID tokens.
        .is_late:
            Is the task late?
        .is_manual_submit:
            Is the latest job submission due to a manual trigger?
        .job_vacated:
            Is the latest job pre-empted (or vacated)?
        .jobs:
            A list of job ids associated with the task proxy.
        .local_job_file_path:
            Path on workflow host to the latest job script for the task.
        .late_time:
            Time in seconds since epoch, beyond which the task is considered
            late if it is never active.
        .non_unique_events (collections.Counter):
            Count non-unique events (e.g. critical, warning, custom).
        .point:
            Cycle point of the task.
        .point_as_seconds:
            Cycle point as seconds since epoch.
        .poll_timer:
            Schedule for polling submitted or running jobs.
        .reload_successor:
            The task proxy object that replaces the current instance on reload.
            This attribute provides a useful link to the latest replacement
            instance while the current object may still be referenced by a job
            manipulation command.
        .submit_num:
            Number of times the task has attempted job submission.
        .summary (dict):
            job_runner_name (str):
                Name of job runner where latest job is submitted.
            description (str):
                Same as the .tdef.rtconfig['meta']['description'] attribute.
            execution_time_limit (float):
                Execution time limit of latest job.
            finished_time (float):
                Latest job exit time.
            finished_time_string (str):
                Latest job exit time as string.
            platforms_used (dict):
                Jobs' platform by submit number.
            label (str):
                The .point attribute as string.
            name (str):
                Same as the .tdef.name attribute.
            started_time (float):
                Latest job execution start time.
            started_time_string (str):
                Latest job execution start time as string.
            submit_method_id (str):
                Latest ID of job in job runner.
            submit_num (int):
                Same as the .submit_num attribute.
            submitted_time (float):
                Latest job submission time.
            submitted_time_string (str):
                Latest job submission time as string.
            title (str):
                Same as the .tdef.rtconfig['meta']['title'] attribute.
        .state:
            Object representing the state of this task.
        .platform:
            Dict containing info for platform where latest job is submitted.
        .tdef:
            The definition object of this task.
        .timeout:
            Timeout value in seconds since epoch for latest job
            submission/execution.
        .try_timers:
            Retry schedules as cylc.flow.task_action_timer.TaskActionTimer
            objects.
        .graph_children (dict)
            graph children: {msg: [(name, point), ...]}
        .flow_nums:
            flows I belong to
         flow_wait:
            wait for flow merge before spawning children
        .waiting_on_job_prep:
            True whilst task is awaiting job prep, reset to False once the
            preparation has completed.
        .transient:
            This is a transient proxy - not to be added to the task pool, but
            used e.g. to spawn children, or to get task-specific information.
        .is_xtrigger_sequential:
            A flag used to determine whether this task needs to wait for
            xtrigger satisfaction to spawn.

    Args:
        tdef: The definition object of this task.
        start_point: Start point to calculate the task's cycle point on
            start-up or the cycle point for subsequent tasks.
        flow_nums: Which flows this task belongs to.
        status: Task state string.
        is_held: True if the task is held, else False.
        submit_num: Number of times the task has attempted job submission.
        is_late: Is the task late?
        data_mode: Reduced store reference data.
    """

    # Memory optimization - constrain possible attributes to this list.
    __slots__ = [
        'clock_trigger_times',
        'expire_time',
        'identity',
        'is_late',
        'is_manual_submit',
        'job_vacated',
        'jobs',
        'late_time',
        'local_job_file_path',
        'non_unique_events',
        'point',
        'point_as_seconds',
        'poll_timer',
        'reload_successor',
        'submit_num',
        'tdef',
        'state',
        'summary',
        'flow_nums',
        'flow_wait',
        'graph_children',
        'platform',
        'timeout',
        'tokens',
        'try_timers',
        'waiting_on_job_prep',
        'mode_settings',
        'transient',
        'is_xtrigger_sequential',
    ]

    def __init__(
        self,
        scheduler_tokens: 'Tokens',
        tdef: 'TaskDef',
        start_point: 'PointBase',
        flow_nums: Optional[Set[int]] = None,
        status: str = TASK_STATUS_WAITING,
        is_held: bool = False,
        submit_num: int = 0,
        is_late: bool = False,
        is_manual_submit: bool = False,
        flow_wait: bool = False,
        data_mode: bool = False,
        transient: bool = False,
        sequential_xtrigger_labels: Optional[Set[str]] = None,
    ) -> None:

        self.tdef = tdef
        if submit_num is None:
            submit_num = 0
        self.submit_num = submit_num
        self.jobs: List[dict] = []
        if flow_nums is None:
            self.flow_nums = set()
        else:
            # (don't share flow_nums ref with parent task)
            self.flow_nums = copy(flow_nums)
        self.flow_wait = flow_wait
        self.point = start_point
        self.tokens = scheduler_tokens.duplicate(
            cycle=str(self.point),
            task=self.tdef.name,
        )
        self.identity = self.tokens.relative_id
        self.reload_successor: Optional['TaskProxy'] = None
        self.point_as_seconds: Optional[int] = None

        self.is_manual_submit = is_manual_submit
        self.summary: Dict[str, Any] = {
            'submitted_time': None,
            'submitted_time_string': None,
            'started_time': None,
            'started_time_string': None,
            'finished_time': None,
            'finished_time_string': None,
            'platforms_used': {},
            'execution_time_limit': None,
            'job_runner_name': None,
            'submit_method_id': None,
            'flow_nums': set(),
            'flow_wait': self.flow_wait
        }

        self.local_job_file_path: Optional[str] = None

        if data_mode:
            self.platform = {}
        else:
            self.platform = get_platform()

        self.transient = transient

        self.job_vacated = False
        self.poll_timer: Optional['TaskActionTimer'] = None
        self.timeout: Optional[float] = None
        self.try_timers: Dict[str, 'TaskActionTimer'] = {}
        self.non_unique_events: TypingCounter[str] = Counter()

        self.clock_trigger_times: Dict[str, int] = {}
        self.expire_time: Optional[float] = None
        self.late_time: Optional[float] = None
        self.is_late = is_late
        self.waiting_on_job_prep = False

        self.state = TaskState(tdef, self.point, status, is_held)

        # Set xtrigger checking type, which effects parentless spawning.
        self.is_xtrigger_sequential = bool(
            sequential_xtrigger_labels
            and self.tdef.is_parentless(start_point)
            and sequential_xtrigger_labels.intersection(self.state.xtriggers)
        )

        # Determine graph children of this task (for spawning).
        if data_mode:
            self.graph_children = {}
        else:
            self.graph_children = generate_graph_children(tdef, self.point)

        self.mode_settings: Optional['ModeSettings'] = None

        if self.tdef.expiration_offset is not None:
            self.expire_time = (
                self.get_point_as_seconds() +
                self.get_offset_as_seconds(
                    self.tdef.expiration_offset
                )
            )

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} '{self.tokens}'>"

    def __str__(self) -> str:
        """Stringify with tokens, state, submit_num, and flow_nums.

        Don't print submit number for pre job-prep states.

        Format: "<point>/<name>/<job>{<flows>}:status".
        """
        id_ = self.identity
        if self.transient:
            return f"{id_}{stringify_flow_nums(self.flow_nums)}"
        if not self.state(TASK_STATUS_WAITING, TASK_STATUS_EXPIRED):
            id_ += f"/{self.submit_num:02d}"
        return (
            f"{id_}{stringify_flow_nums(self.flow_nums)}:{self.state}"
        )

    def copy_to_reload_successor(self, reload_successor, check_output):
        """Copy attributes to successor on reload of this task proxy."""
        self.reload_successor = reload_successor
        reload_successor.submit_num = self.submit_num
        reload_successor.flow_wait = self.flow_wait
        reload_successor.is_manual_submit = self.is_manual_submit
        reload_successor.summary = self.summary
        reload_successor.local_job_file_path = self.local_job_file_path
        reload_successor.try_timers = self.try_timers
        reload_successor.platform = self.platform
        reload_successor.job_vacated = self.job_vacated
        reload_successor.poll_timer = self.poll_timer
        reload_successor.timeout = self.timeout
        reload_successor.state.outputs = self.state.outputs
        reload_successor.state.is_held = self.state.is_held
        reload_successor.state.is_runahead = self.state.is_runahead
        reload_successor.state.is_updated = self.state.is_updated
        reload_successor.mode_settings = self.mode_settings

        # Prerequisites: the graph might have changed before reload, so
        # we need to use the new prerequisites but update them with the
        # pre-reload state of prerequisites that still exist post-reload.

        # Get all prereq states, e.g. {('1', 'c', 'succeeded'): False, ...}
        pre_reload = {
            k: v
            for pre in self.state.prerequisites
            for (k, v) in pre.satisfied.items()
        }
        # Use them to update the new prerequisites.
        # - unchanged prerequisites will keep their pre-reload state.
        # - removed prerequisites will not be carried over
        # - added prerequisites will be recorded as unsatisfied
        #   NOTE: even if the corresponding output was completed pre-reload!
        for pre in reload_successor.state.prerequisites:
            for k in pre.satisfied.keys():
                try:
                    pre.satisfied[k] = pre_reload[k]
                except KeyError:
                    # Look through task outputs to see if is has been
                    # satisfied
                    pre.satisfied[k] = check_output(
                        *k,
                        self.flow_nums,
                    )

        reload_successor.state.xtriggers.update({
            # copy across any special "_cylc" xtriggers which were added
            # dynamically at runtime (i.e. execution retry xtriggers)
            key: value
            for key, value in self.state.xtriggers.items()
            if key.startswith('_cylc')
        })
        reload_successor.jobs = self.jobs

    @staticmethod
    def get_offset_as_seconds(offset):
        """Return an ISO interval as seconds."""
        iso_offset = interval_parse(str(offset))
        return int(iso_offset.get_seconds())

    def get_late_time(self):
        """Compute and store late time as seconds since epoch."""
        if self.late_time is None:
            if self.tdef.rtconfig['events']['late offset']:
                self.late_time = (
                    self.get_point_as_seconds() +
                    self.tdef.rtconfig['events']['late offset'])
            else:
                # Not used, but allow skip of the above "is None" test
                self.late_time = 0
        return self.late_time

    def get_point_as_seconds(self):
        """Compute and store my cycle point as seconds since epoch."""
        if self.point_as_seconds is None:
            iso_timepoint = point_parse(str(self.point))
            self.point_as_seconds = int(iso_timepoint.seconds_since_unix_epoch)
            if iso_timepoint.time_zone.unknown:
                utc_offset_hours, utc_offset_minutes = (
                    get_local_time_zone())
                utc_offset_in_seconds = (
                    3600 * utc_offset_hours + 60 * utc_offset_minutes)
                self.point_as_seconds += utc_offset_in_seconds
        return self.point_as_seconds

    def get_clock_trigger_time(
        self,
        point: 'PointBase', offset_str: Optional[str] = None
    ) -> int:
        """Compute, cache and return trigger time relative to cycle point.

        Args:
            point: Task's cycle point.
            offset_str: ISO8601 interval string, e.g. "PT2M".
                Can be None for zero offset.
        Returns:
            Absolute trigger time in seconds since Unix epoch.

        """
        offset_str = offset_str if offset_str else 'P0Y'
        if offset_str not in self.clock_trigger_times:
            if offset_str == 'P0Y':
                trigger_time = point
            else:
                trigger_time = point + ISO8601Interval(offset_str)

            offset = int(
                point_parse(str(trigger_time)).seconds_since_unix_epoch)
            self.clock_trigger_times[offset_str] = offset
        return self.clock_trigger_times[offset_str]

    def get_try_num(self):
        """Return the number of automatic tries (try number)."""
        try:
            return self.try_timers[TimerFlags.EXECUTION_RETRY].num + 1
        except (AttributeError, KeyError):
            return 0

    def next_point(self):
        """Return the next cycle point."""
        return self.tdef.next_point(self.point)

    def is_ready_to_run(self) -> Tuple[bool, ...]:
        """Is this task ready to run?

        Takes account of all dependence: on other tasks, xtriggers, and
        old-style ext-triggers. Or, manual triggering.

        """
        if self.is_manual_submit:
            # Manually triggered, ignore unsatisfied prerequisites.
            return (True,)
        if self.state.is_held:
            # A held task is not ready to run.
            return (False,)
        if self.state.status in self.try_timers:
            # A try timer is still active.
            return (self.try_timers[self.state.status].is_delay_done(),)
        return (
            self.state(TASK_STATUS_WAITING),
            self.is_waiting_prereqs_done()
        )

    def set_summary_time(self, event_key, time_str=None):
        """Set an event time in self.summary

        Set values of both event_key + "_time" and event_key + "_time_string".
        """
        if time_str is None:
            self.summary[event_key + '_time'] = None
        else:
            self.summary[event_key + '_time'] = float(str2time(time_str))
        self.summary[event_key + '_time_string'] = time_str

    def is_task_prereqs_not_done(self):
        """Are some task prerequisites not satisfied?"""
        return (not all(pre.is_satisfied()
                for pre in self.state.prerequisites))

    def is_waiting_prereqs_done(self):
        """Are ALL prerequisites satisfied?"""
        return (
            all(pre.is_satisfied() for pre in self.state.prerequisites)
            and self.state.external_triggers_all_satisfied()
            and self.state.xtriggers_all_satisfied()
        )

    def reset_try_timers(self):
        # unset any retry delay timers
        for timer in self.try_timers.values():
            timer.timeout = None

    def status_match(self, status: Optional[str]) -> bool:
        """Return whether a string matches the task's status.

        None/an empty string is treated as a match.
        """
        return (not status) or self.state.status == status

    def name_match(
        self,
        value: str,
        match_func: Callable[[Any, Any], bool] = fnmatchcase
    ) -> bool:
        """Return whether a string/pattern matches the task's name or any of
        its parent family names."""
        return match_func(self.tdef.name, value) or any(
            match_func(ns, value) for ns in self.tdef.namespace_hierarchy
        )

    def merge_flows(self, flow_nums: Set) -> None:
        """Merge another set of flow_nums with mine."""
        self.flow_nums.update(flow_nums)
        LOG.info(
            f"[{self}] merged in flow(s) "
            f"{','.join(str(f) for f in flow_nums)}"
        )

    def state_reset(
        self, status=None, is_held=None, is_queued=None, is_runahead=None,
        silent=False, forced=False
    ) -> bool:
        """Set new state and log the change. Return whether it changed.

        """
        before = str(self)

        if status == TASK_STATUS_EXPIRED:
            is_queued = False
            is_runahead = False

        if self.state.reset(
            status, is_held, is_queued, is_runahead, forced
        ):
            if not silent and not self.transient:
                LOG.info(f"[{before}] => {self.state}")
            return True

        return False

    def satisfy_me(
        self, task_messages: 'List[Tokens]'
    ) -> 'Set[Tokens]':
        """Try to satisfy my prerequisites with given output messages.

        The task output messages are of the form "cycle/task:message"
        Log a warning for messages that I don't depend on.

        Return a set of unmatched task messages.

        """
        used = self.state.satisfy_me(task_messages)
        return set(task_messages) - used

    def clock_expire(self) -> bool:
        """Return True if clock expire time is up, else False."""
        if (
            self.expire_time is None  # expiry not configured
            or self.state(TASK_STATUS_EXPIRED)  # already expired
            or time() < self.expire_time  # not time yet
        ):
            return False
        return True

    def is_complete(self) -> bool:
        """Return True if complete or expired."""
        return self.state.outputs.is_complete()



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""cylc support for the Jinja2 template processor

Importing code should catch ImportError in case Jinja2 is not installed.
"""

from contextlib import suppress
from glob import glob
import importlib
import os
import pkgutil
import re
import sys
import traceback
import typing as t

from jinja2 import (
    BaseLoader,
    ChoiceLoader,
    Environment,
    FileSystemLoader,
    StrictUndefined,
    TemplateNotFound,
    TemplateSyntaxError)

from cylc.flow import LOG
from cylc.flow.parsec.exceptions import Jinja2Error
from cylc.flow.parsec.fileparse import get_cylc_env_vars

TRACEBACK_LINENO = re.compile(
    r'\s+?File "(?P<file>.*)", line (?P<line>\d+), in .*template'
)
CONTEXT_LINES = 3


class PyModuleLoader(BaseLoader):
    """Load python module as Jinja2 template.

    This loader piggybacks on the jinja import mechanism and
    returns an empty template that exports module's namespace."""

    # no source access for this loader
    has_source_access = False

    def __init__(self, prefix='__python__'):
        self._templates = {}
        # prefix that can be used to avoid name collisions with template files
        self._python_namespace_prefix = prefix + '.'

    def load(
        self,
        environment,
        name,
        globals=None  # noqa: A002 (required to match underlying interface?)
    ):
        """Imports Python module and returns it as Jinja2 template."""
        if name.startswith(self._python_namespace_prefix):
            name = name[len(self._python_namespace_prefix):]
        with suppress(KeyError):
            return self._templates[name]
        try:
            mdict = __import__(name, fromlist=['*']).__dict__
        except ImportError:
            raise TemplateNotFound(name)

        # inject module dict into the context of an empty template
        def root_render_func(context, *args, **kwargs):
            """Template render function."""
            if False:
                yield None  # to make it a generator
            context.vars.update(mdict)
            context.exported_vars.update(mdict)

        templ = environment.from_string('')
        templ.root_render_func = root_render_func
        self._templates[name] = templ
        return templ


def raise_helper(message, error_type='Error'):
    """Provides a Jinja2 function for raising exceptions."""
    # TODO - this more nicely
    raise Exception('Jinja2 %s: %s' % (error_type, message))


def assert_helper(logical, message):
    """Provides a Jinja2 function for asserting logical expressions."""
    if not logical:
        raise_helper(message, 'Assertion Error')
    return ''  # Prevent None return value polluting output.


def _load_jinja2_extensions():
    """
    Load modules under the cylc.jinja package namespace.

    Filters provided by third-party packages (i.e. user created packages) will
    also be included if correctly put in the cylc.jinja.filters namespace.

    Global variables are expected to be found in cylc.jinja.globals,
    and jinja tests in cylc.jinja.tests.

    The dictionary returned contains the full module name (e.g.
    cylc.jinja.filters.pad), and the second value is the module
    object (same object as in __import__("module_name")__).

    :return: jinja2 filter modules
    :rtype: dict[string, object]
    """
    jinja2_extensions = {}
    for module_name in [
        "cylc.flow.jinja.filters",
        "cylc.flow.jinja.globals",
        "cylc.flow.jinja.tests"
    ]:
        try:
            module = importlib.import_module(module_name)
            jinja2_filters_modules = pkgutil.iter_modules(
                module.__path__, f"{module.__name__}.")
            if jinja2_filters_modules:
                namespace = module_name.split(".")[-1]
                jinja2_extensions[namespace] = {
                    name.split(".")[-1]: importlib.import_module(name)
                    for finder, name, ispkg in jinja2_filters_modules
                }
        except ModuleNotFoundError:
            # Nothing to do, we may start without any filters/globals/tests
            pass

    return jinja2_extensions


def jinja2environment(dir_=None):
    """Set up and return Jinja2 environment."""
    if dir_ is None:
        dir_ = os.getcwd()

    # Ignore bandit false positive: B701:jinja2_autoescape_false
    # This env is not used to render content that is vulnerable to XSS.
    env = Environment(  # nosec
        loader=ChoiceLoader([FileSystemLoader(dir_), PyModuleLoader()]),
        undefined=StrictUndefined,
        extensions=['jinja2.ext.do'])

    # Load Jinja2 filters using setuptools
    for scope, extensions in _load_jinja2_extensions().items():
        for fname, module in extensions.items():
            getattr(env, scope)[fname] = getattr(module, fname)

    # Load any custom Jinja2 filters, tests or globals in the workflow
    # definition directory
    # Example: a filter to pad integer values some fill character:
    # |(file WORKFLOW_DEFINITION_DIRECTORY/Jinja2/foo.py)
    # |  #!/usr/bin/env python3
    # |  def foo( value, length, fillchar ):
    # |     return str(value).rjust( int(length), str(fillchar) )
    for namespace in ['filters', 'tests', 'globals']:
        nspdir = 'Jinja2' + namespace.capitalize()
        fdirs = [os.path.join(dir_, nspdir)]
        try:
            fdirs.append(os.path.join(os.environ['HOME'], '.cylc', nspdir))
        except KeyError:
            # (Needed for tests/f/cylc-get-site-config/04-homeless.t!)
            LOG.warning(f"$HOME undefined: can't load ~/.cylc/{nspdir}")
        for fdir in fdirs:
            if os.path.isdir(fdir):
                sys.path.insert(1, os.path.abspath(fdir))
                for name in glob(os.path.join(fdir, '*.py')):
                    fname = os.path.splitext(os.path.basename(name))[0]
                    # TODO - EXCEPTION HANDLING FOR LOADING CUSTOM FILTERS
                    module = __import__(fname)
                    envnsp = getattr(env, namespace)
                    envnsp[fname] = getattr(module, fname)

    # Import WORKFLOW HOST USER ENVIRONMENT into template:
    # (Usage e.g.: {{environ['HOME']}}).
    env.globals['environ'] = os.environ
    env.globals['raise'] = raise_helper
    env.globals['assert'] = assert_helper

    # Add `CYLC_` environment variables to the global namespace.
    env.globals.update(
        get_cylc_env_vars()
    )
    return env


def get_error_lines(
    base_template_file: str,
    template_lines: t.List[str],
) -> t.Dict[str, t.List[str]]:
    """Extract exception lines from Jinja2 tracebacks.

    Returns:
        {filename: [exception_line, ...]}

        There may be multiple entries due to {% include %} statements.

    """
    ret = {}
    for line in reversed(traceback.format_exc().splitlines()):
        match = TRACEBACK_LINENO.match(line)
        lines: t.List[str] = []
        if match:
            filename = match.groupdict()['file']
            lineno = int(match.groupdict()['line'])
            start_line = max(lineno - CONTEXT_LINES, 0)
            if filename in {'<template>', '<unknown>'}:
                filename = base_template_file
                lineno += 1  # shebang line ignored by jinja2
                lines = template_lines[start_line:lineno]
            elif os.path.isfile(filename):
                with open(filename, 'r') as jinja2_file:
                    for i, fline in enumerate(jinja2_file, start=1):
                        if i < start_line:
                            continue
                        if i > lineno:
                            break
                        # use splitlines to remove the newline char at the
                        # end of the line
                        lines.append(fline.splitlines()[0])
            if lines:
                ret[filename] = lines

    return ret


def jinja2process(
    fpath: str,
    flines: t.List[str],
    dir_: str,
    template_vars: t.Optional[t.Dict[str, t.Any]] = None,
) -> t.List[str]:
    """Pass configure file through Jinja2 processor.

    Args:
        fpath:
            The path to the root template file (i.e. the flow.cylc file)
        flines:
            List of template lines to process.
        dir_:
            The path to the configuration directory.
        template_vars:
            Dictionary of template variables.

    """
    # Load file lines into a template, excluding '#!jinja2' so that
    # '#!cylc-x.y.z' rises to the top. Callers should handle jinja2
    # TemplateSyntaxerror and TemplateError.
    if template_vars:
        LOG.debug(
            'Setting Jinja2 template variables:\n%s',
            '\n'.join(
                ['+ %s=%s' % item for item in sorted(template_vars.items())]))

    # Jinja2 render method requires a dictionary as argument (not None):
    if not template_vars:
        template_vars = {}

    # CALLERS SHOULD HANDLE JINJA2 TEMPLATESYNTAXERROR AND TEMPLATEERROR
    # AND TYPEERROR (e.g. for not using "|int" filter on number inputs.
    # Convert unicode to plain str, ToDo - still needed for parsec?)
    try:
        env = jinja2environment(dir_)
        template = env.from_string('\n'.join(flines[1:]))
        lines = str(template.render(template_vars)).splitlines()
    except TemplateSyntaxError as exc:
        filename = None
        # extract source lines
        if exc.lineno and exc.source and not exc.filename:
            # error in flow.cylc or cylc include file
            lines = exc.source.splitlines()
        elif exc.lineno and exc.filename:
            # error in jinja2 include file
            filename = os.path.relpath(exc.filename, dir_)
            with open(exc.filename, 'r') as include_file:
                include_file.seek(max(exc.lineno - CONTEXT_LINES, 0), 0)
                lines = []
                for _ in range(CONTEXT_LINES):
                    lines.append(include_file.readline().splitlines()[0])

        raise Jinja2Error(
            exc,
            lines=get_error_lines(fpath, flines),
            filename=filename
        )
    except Exception as exc:
        raise Jinja2Error(
            exc,
            lines=get_error_lines(fpath, flines),
        )

    # Ignore blank lines (lone Jinja2 statements leave blank lines behind)
    return [line for line in lines if line.strip()]



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""cylc get-workflow-contact [OPTIONS] ARGS

Print contact information of a running workflow.
"""

from typing import TYPE_CHECKING

from cylc.flow.exceptions import CylcError, ServiceFileError
from cylc.flow.id_cli import parse_id
from cylc.flow.option_parsers import (
    WORKFLOW_ID_ARG_DOC,
    CylcOptionParser as COP,
)
from cylc.flow.workflow_files import load_contact_file
from cylc.flow.terminal import cli_function

if TYPE_CHECKING:
    from optparse import Values


def get_option_parser():
    return COP(__doc__, argdoc=[WORKFLOW_ID_ARG_DOC])


@cli_function(get_option_parser)
def main(parser: COP, options: 'Values', workflow_id: str) -> None:
    """CLI for "cylc get-workflow-contact"."""
    workflow_id, *_ = parse_id(
        workflow_id,
        constraint='workflows',
    )
    try:
        data = load_contact_file(workflow_id)
    except ServiceFileError:
        raise CylcError(
            f"{workflow_id}: cannot get contact info, workflow not running?"
        )
    else:
        for key, value in sorted(data.items()):
            print("%s=%s" % (key, value))



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from functools import partial
from subprocess import Popen, PIPE

from cylc.flow.exceptions import (
    ClientError,
    ClientTimeout,
    WorkflowStopped,
)
from cylc.flow.network.client_factory import get_client
from cylc.flow.id import Tokens
from cylc.flow.tui.util import (
    extract_context
)


# the GraphQL query which Tui runs against each of the workflows
# is is subscribed to
QUERY = '''
  query cli($taskStates: [String]){
    workflows {
      id
      name
      port
      status
      stateTotals
      taskProxies(states: $taskStates) {
        id
        name
        cyclePoint
        state
        isHeld
        isQueued
        isRunahead
        firstParent {
          id
          name
        }
        jobs(sort: { keys: ["submit_num"], reverse: true}) {
          id
          submitNum
          state
          platform
          jobRunnerName
          jobId
          startedTime
          finishedTime
        }
        task {
          meanElapsedTime
        }
      }
      familyProxies(exids: ["*/root"], states: $taskStates) {
        id
        name
        cyclePoint
        state
        isHeld
        isQueued
        isRunahead
        firstParent {
          id
          name
        }
      }
      cyclePoints: familyProxies(ids: ["*/root"], states: $taskStates) {
        id
        cyclePoint
        state
        isHeld
        isQueued
        isRunahead
      }
    }
  }
'''

# the list of mutations we can call on a running scheduler
MUTATIONS = {
    'workflow': [
        'pause',
        'reload',
        'stop',
    ],
    'cycle': [
        'hold',
        'release',
        'kill',
        'trigger',
        'poll',
    ],
    'task': [
        'hold',
        'release',
        'kill',
        'trigger',
        'poll',
        'set',
    ],
    'job': [
        'kill',
    ]
}

# mapping of Tui's node types (e.g. workflow) onto GraphQL argument types
# (e.g. WorkflowID)
ARGUMENT_TYPES = {
    # <tui-node-type>: <graphql-argument-type>
    'workflow': '[WorkflowID]!',
    'task': '[NamespaceIDGlob]!',
}


def cli_cmd(*cmd, ret=False):
    """Issue a CLI command.

    Args:
        cmd:
            The command without the 'cylc' prefix'.
        ret:
            If True, the stdout will be returned.

    Rasies:
        ClientError:
            In the event of mishap for consistency with the network
            client alternative.

    """
    proc = Popen(  # nosec (command constructed internally, no untrusted input)
        ['cylc', *cmd],
        stderr=PIPE,
        stdout=PIPE,
        text=True,
    )
    _out, err = proc.communicate()
    if proc.returncode != 0:
        raise ClientError(f'Error in command cylc {" ".join(cmd)}\n{err}')
    if ret:
        return _out


def _show(id_):
    """Special mutation to display cylc show output."""
    # dynamic import to avoid circular import issues
    from cylc.flow.tui.overlay import text_box
    return partial(
        text_box,
        text=cli_cmd('show', id_, '--color=never', ret=True),
    )


def _log(id_):
    """Special mutation to open the log view."""
    # dynamic import to avoid circular import issues
    from cylc.flow.tui.overlay import log
    return partial(
        log,
        id_=id_,
        list_files=partial(_list_log_files, id_),
        get_log=partial(_get_log, id_),
    )


def _parse_log_header(contents):
    """Parse the cat-log header.

    The "--prepend-path" option to "cat-log" adds a line containing the host
    and path to the file being viewed in the form:

        # <host>:<path>

    Args:
        contents:
            The raw log file contents as returned by "cat-log".

    Returns:
        tuple - (host, path, text)

        host:
            The host where the file was retrieved from.
        path:
            The absolute path to the log file.
        text:
            The log file contents with the header removed.

    """
    contents, text = contents.split('\n', 1)
    contents = contents.replace('# ', '')
    host, path = contents.split(':')
    return host, path, text


def _get_log(id_, filename=None):
    """Retrieve the contents of a log file.

    Args:
        id_:
            The Cylc universal ID of the thing you want to fetch the log file
            for.
        filename:
            The file name to retrieve (note name not path).
            If "None", then the default log file will be retrieved.

    """
    cmd = [
        'cat-log',
        '--mode=cat',
        '--prepend-path',
    ]
    if filename:
        cmd.append(f'--file={filename}')
    text = cli_cmd(
        *cmd,
        id_,
        ret=True,
    )
    return _parse_log_header(text)


def _list_log_files(id_):
    """Return a list of available log files.

    Args:
        id_:
            The Cylc universal ID of the thing you want to fetch the log file
            for.

    """
    text = cli_cmd('cat-log', '--mode=list-dir', id_, ret=True)
    return text.splitlines()


# the mutations we have to go through the CLI to perform
OFFLINE_MUTATIONS = {
    'user': {
        'stop-all': partial(cli_cmd, 'stop', '*'),
    },
    'workflow': {
        'play': partial(cli_cmd, 'play'),
        'clean': partial(cli_cmd, 'clean', '--yes'),
        'reinstall-reload': partial(cli_cmd, 'vr', '--yes'),
        'log': _log,
    },
    'task': {
        'log': _log,
        'show': _show,
    },
    'job': {
        'log': _log,
    },
}


def generate_mutation(mutation, arguments):
    """Return a GraphQL mutation string.

    Args:
        mutation:
            The mutation name.
        Arguments:
            The arguments to provide to it.

    """
    arguments.pop('user')
    graphql_args = ', '.join([
        f'${argument}: {ARGUMENT_TYPES[argument]}'
        for argument in arguments
    ])
    mutation_args = ', '.join([
        f'{argument}s: ${argument}'
        for argument in arguments
    ])

    return f'''
        mutation({graphql_args}) {{
            {mutation} ({mutation_args}) {{
                result
            }}
        }}
    '''


def list_mutations(selection, is_running=True):
    """List mutations relevant to the provided selection.

    Args:
        selection:
            The user selection.
        is_running:
            If False, then mutations which require the scheduler to be
            running will be omitted.

            Note, this is only relevant for workflow nodes because if a
            workflow is stopped, then any tasks within it will be removed
            anyway.

    """
    context = extract_context(selection)
    selection_type = list(context)[-1]
    ret = []
    if is_running:
        # add the online mutations
        ret.extend(MUTATIONS.get(selection_type, []))
    # add the offline mutations
    ret.extend(OFFLINE_MUTATIONS.get(selection_type, []))
    return sorted(ret)


def context_to_variables(context, jobs=False):
    """Derive multiple selection out of single selection.

    Note, this interface exists with the aim of facilitating the addition of
    multiple selection at a later date.

    Examples:
        >>> context_to_variables(extract_context(['~a/b//c/d']))
        {'user': ['a'], 'workflow': ['b'], 'task': ['c/d']}

        >>> context_to_variables(extract_context(['~a/b//c']))
        {'user': ['a'], 'workflow': ['b'], 'task': ['c/*']}

        >>> context_to_variables(extract_context(['~a/b']))
        {'user': ['a'], 'workflow': ['b']}

        # Note, jobs are omitted by default
        >>> context_to_variables(extract_context(['~a/b//c/d/01']))
        {'user': ['a'], 'workflow': ['b'], 'task': ['c/d']}

        # This is because Cylc commands cannot generally operate on jobs only
        # tasks.
        # To let jobs slide through:
        >>> context_to_variables(extract_context(['~a/b//c/d/01']), jobs=True)
        {'user': ['a'], 'workflow': ['b'], 'job': ['c/d/01']}

    """
    # context_to_variables because it can only handle single-selection ATM
    variables = {'user': context['user']}

    if 'workflow' in context:
        variables['workflow'] = context['workflow']
    if jobs and 'job' in context:
        variables['job'] = [
            Tokens(
                cycle=context['cycle'][0],
                task=context['task'][0],
                job=context['job'][0],
            ).relative_id
        ]
    elif 'task' in context:
        variables['task'] = [
            Tokens(
                cycle=context['cycle'][0],
                task=context['task'][0]
            ).relative_id
        ]
    elif 'cycle' in context:
        variables['task'] = [
            Tokens(cycle=context['cycle'][0], task='*').relative_id
        ]
    return variables


def mutate(mutation, selection):
    """Call a mutation.

    Args:
        mutation:
            The mutation name (e.g. stop).
        selection:
            The Tui selection (i.e. the row(s) selected in Tui).

    """
    if mutation in {
        _mutation
        for section in OFFLINE_MUTATIONS.values()
        for _mutation in section
    }:
        return offline_mutate(mutation, selection)
    else:
        online_mutate(mutation, selection)
    return None


def online_mutate(mutation, selection):
    """Issue a mutation over a network interface."""
    context = extract_context(selection)
    variables = context_to_variables(context)

    # note this only supports single workflow mutations at present
    workflow = variables['workflow'][0]
    try:
        client = get_client(workflow)
    except WorkflowStopped:
        raise Exception(
            f'Cannot peform command {mutation} on a stopped workflow'
        )
    except (ClientError, ClientTimeout) as exc:
        raise Exception(
            f'Error connecting to workflow: {exc}'
        )

    request_string = generate_mutation(mutation, variables)
    client(
        'graphql',
        {
            'request_string': request_string,
            'variables': variables
        }
    )


def offline_mutate(mutation, selection):
    """Issue a mutation over the CLI or other offline interface."""
    context = extract_context(selection)
    variables = context_to_variables(context, jobs=True)
    if 'job' in variables:
        for job in variables['job']:
            id_ = Tokens(job, relative=True).duplicate(
                workflow=variables['workflow'][0]
            )
            return OFFLINE_MUTATIONS['job'][mutation](id_.id)
    if 'task' in variables:
        for task in variables['task']:
            id_ = Tokens(task, relative=True).duplicate(
                workflow=variables['workflow'][0]
            )
            return OFFLINE_MUTATIONS['task'][mutation](id_.id)
    if 'workflow' in variables:
        for workflow in variables['workflow']:
            return OFFLINE_MUTATIONS['workflow'][mutation](workflow)
    else:
        return OFFLINE_MUTATIONS['user'][mutation]()
